from pathlib import Path
from typing import Any, List

from sbibm.tasks.task import Task


def get_task(task_name: str, *args: Any, **kwargs: Any) -> Task:
    """Get task

    Args:
        task_name: Name of task

    Returns:
        Task instance
    """
    if task_name == "lotka_volterra":
        from sbibm.tasks.lotka_volterra.task import LotkaVolterra

        return LotkaVolterra(*args, **kwargs)

    elif task_name == "bernoulli_glm":
        from sbibm.tasks.bernoulli_glm.task import BernoulliGLM

        return BernoulliGLM(*args, **kwargs)

    elif task_name == "bernoulli_glm_raw":
        from sbibm.tasks.bernoulli_glm.task import BernoulliGLM

        return BernoulliGLM(*args, summary="raw", **kwargs)

    elif task_name == "gaussian_linear":
        from sbibm.tasks.gaussian_linear.task import GaussianLinear

        return GaussianLinear(*args, **kwargs)

    elif task_name == "gaussian_linear_uniform":
        from sbibm.tasks.gaussian_linear_uniform.task import GaussianLinearUniform

        return GaussianLinearUniform(*args, **kwargs)

    elif task_name == "gaussian_mixture":
        from sbibm.tasks.gaussian_mixture.task import GaussianMixture

        return GaussianMixture(*args, **kwargs)

    elif task_name == "slcp" or task_name == "gaussian_nonlinear":
        from sbibm.tasks.slcp.task import SLCP

        return SLCP(*args, **kwargs)

    elif task_name == "slcp_distractors":
        from sbibm.tasks.slcp.task import SLCP

        return SLCP(*args, distractors=True, **kwargs)

    if task_name == "sir":
        from sbibm.tasks.sir.task import SIR

        return SIR(*args, **kwargs)

    elif task_name == "two_moons":
        from sbibm.tasks.two_moons.task import TwoMoons

        return TwoMoons(*args, **kwargs)

    else:
        raise NotImplementedError()


def get_task_name_display(task_name: str, *args: Any, **kwargs: Any) -> str:
    return get_task(task_name).name_display


def get_available_tasks() -> List[str]:
    """Get available tasks

    Returns:
        List of tasks
    """
    task_dir = Path(__file__).parent.absolute()
    tasks = [f.name for f in task_dir.glob("*") if f.is_dir() and f.name[0] != "_"]
    tasks_extra = ["slcp_distractors", "bernoulli_glm_raw"]
    return tasks + tasks_extra



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0
import warnings
from collections import OrderedDict
from functools import partial
from typing import Any, Callable, Dict

import pyro
import torch
from opt_einsum import shared_intermediates
from pyro import distributions as dist
from pyro import poutine as poutine
from pyro.infer import config_enumerate
from pyro.infer.enum import iter_discrete_traces
from pyro.infer.util import is_validation_enabled
from pyro.ops.contract import contract_to_tensor
from pyro.poutine.subsample_messenger import _Subsample
from pyro.util import check_site_shape, ignore_jit_warnings
from torch.autograd import grad
from torch.distributions import biject_to


def get_log_prob_fn(
    model,
    model_args=(),
    model_kwargs={},
    implementation="pyro",
    automatic_transform_enabled=False,
    transforms=None,
    max_plate_nesting=None,
    jit_compile=False,
    jit_options=None,
    skip_jit_warnings=False,
    **kwargs,
) -> (Callable, Dict[str, Any]):
    """
    Given a Python callable with Pyro primitives, generates the following model-specific
    functions:
    - a log prob function whose input are parameters and whose output
      is the log prob of the model
    - transforms to transform latent sites of `model` to unconstrained space

    Args:
        model: a Pyro model which contains Pyro primitives.
        model_args: optional args taken by `model`.
        model_kwargs: optional kwargs taken by `model`.
        implementation: Switches between implementations
        automatic_transform_enabled: Whether or not should try to infer transforms
            to unconstrained space
        transforms: Optional dictionary that specifies a transform
            for a sample site with constrained support to unconstrained space. The
            transform should be invertible, and implement `log_abs_det_jacobian`.
            If not specified and the model has sites with constrained support,
            automatic transformations will be applied, as specified in
            `torch.distributions.constraint_registry`.
        max_plate_nesting: Optional bound on max number of nested
            `pyro.plate` contexts. This is required if model contains
            discrete sample sites that can be enumerated over in parallel. Will
            try to infer automatically if not provided
        jit_compile: Optional parameter denoting whether to use
            the PyTorch JIT to trace the log density computation, and use this
            optimized executable trace in the integrator.
        jit_options: A dictionary contains optional arguments for
            `torch.jit.trace` function.
        ignore_jit_warnings: Flag to ignore warnings from the JIT
            tracer when `jit_compile=True`. Default is False.

    Returns:
        `log_prob_fn` and `transforms`
    """
    if transforms is None:
        transforms = {}

    if max_plate_nesting is None:
        max_plate_nesting = _guess_max_plate_nesting(model, model_args, model_kwargs)

    model = poutine.enum(
        config_enumerate(model), first_available_dim=-1 - max_plate_nesting
    )
    model_trace = poutine.trace(model).get_trace(*model_args, **model_kwargs)

    has_enumerable_sites = False
    for name, node in model_trace.iter_stochastic_nodes():
        fn = node["fn"]

        if isinstance(fn, _Subsample):
            if fn.subsample_size is not None and fn.subsample_size < fn.size:
                raise NotImplementedError(
                    "Model with subsample sites are not supported."
                )
            continue

        if fn.has_enumerate_support:
            has_enumerable_sites = True
            continue

        if automatic_transform_enabled:
            transforms[name] = biject_to(fn.support).inv
        else:
            transforms[name] = dist.transforms.identity_transform

    if implementation == "pyro":
        trace_prob_evaluator = TraceEinsumEvaluator(
            model_trace, has_enumerable_sites, max_plate_nesting
        )

        lp_maker = _LPMaker(
            model, model_args, model_kwargs, trace_prob_evaluator, transforms
        )

        lp_fn = lp_maker.get_lp_fn(jit_compile, skip_jit_warnings, jit_options)

    elif implementation == "experimental":
        assert automatic_transform_enabled is False

        if jit_compile:
            warnings.warn("Will not JIT compile, unsupported for now.")

        def lp_fn(input_dict):
            excluded_nodes = set(["_INPUT", "_RETURN"])

            for key, value in input_dict.items():
                model_trace.nodes[key]["value"] = value

            replayed_model = pyro.poutine.replay(model, model_trace)

            log_p = 0
            for trace_enum in iter_discrete_traces("flat", fn=replayed_model):
                trace_enum.compute_log_prob()

                for node_name, node in trace_enum.nodes.items():
                    if node_name in excluded_nodes:
                        continue

                    if node["log_prob"].ndim == 1:
                        log_p += trace_enum.nodes[node_name]["log_prob"]
                    else:
                        log_p += trace_enum.nodes[node_name]["log_prob"].sum(dim=1)

            return log_p

    else:
        raise NotImplementedError

    return lp_fn, transforms


def get_log_prob_grad_fn(
    model,
    model_args=(),
    model_kwargs={},
    implementation="pyro",
    automatic_transform_enabled=False,
    transforms=None,
    max_plate_nesting=None,
    jit_compile=False,
    jit_options=None,
    skip_jit_warnings=False,
    **kwargs,
) -> (Callable, Dict[str, Any]):
    """
    Given a Python callable with Pyro primitives, generates the following model-specific
    functions:
    - a log prob grad function whose input are parameters and whose 
      output is the grd of log prob of the model wrt parameters
    - transforms to transform latent sites of `model` to 
      unconstrained space

    Args:
        See `get_log_prob_fn`

    Returns:
        `log_prob_grad_fn` and `transforms`
    """
    lp_fn, transforms = get_log_prob_fn(
        model,
        model_args,
        model_kwargs,
        implementation,
        automatic_transform_enabled,
        transforms,
        max_plate_nesting,
        jit_compile,
        jit_options,
        skip_jit_warnings,
    )
    lp_grad_fn = make_log_prob_grad_fn(lp_fn)
    return lp_grad_fn, transforms


class _LPMaker:
    def __init__(
        self, model, model_args, model_kwargs, trace_prob_evaluator, transforms
    ):
        self.model = model
        self.model_args = model_args
        self.model_kwargs = model_kwargs
        self.trace_prob_evaluator = trace_prob_evaluator
        self.transforms = transforms
        self._compiled_fn = None

    def _lp_fn(self, params):
        params_constrained = {k: self.transforms[k].inv(v) for k, v in params.items()}
        cond_model = poutine.condition(self.model, params_constrained)
        model_trace = poutine.trace(cond_model).get_trace(
            *self.model_args, **self.model_kwargs
        )
        log_joint = self.trace_prob_evaluator.log_prob(model_trace)
        for name, t in self.transforms.items():
            log_joint = log_joint - torch.sum(
                t.log_abs_det_jacobian(params_constrained[name], params[name])
            )
        return log_joint

    def _lp_fn_jit(self, skip_jit_warnings, jit_options, params):
        if not params:
            return self._lp_fn(params)
        names, vals = zip(*sorted(params.items()))

        if self._compiled_fn:
            return self._compiled_fn(*vals)

        with pyro.validation_enabled(False):
            tmp = []
            for _, v in pyro.get_param_store().named_parameters():
                if v.requires_grad:
                    v.requires_grad_(False)
                    tmp.append(v)

            def _lp_jit(*zi):
                params = dict(zip(names, zi))
                return self._lp_fn(params)

            if skip_jit_warnings:
                _lp_jit = ignore_jit_warnings()(_lp_jit)
            self._compiled_fn = torch.jit.trace(_lp_jit, vals, **jit_options)

            for v in tmp:
                v.requires_grad_(True)
            return self._compiled_fn(*vals)

    def get_lp_fn(self, jit_compile=False, skip_jit_warnings=True, jit_options=None):
        if jit_compile:
            jit_options = {"check_trace": False} if jit_options is None else jit_options
            return partial(self._lp_fn_jit, skip_jit_warnings, jit_options)
        return self._lp_fn


def make_log_prob_grad_fn(log_prob_fn):
    """Makes `log_prob_grad_fn`

    Args:
        log_prob_fn: python callable that takes in a dictionary of parameters
        and returns the log prob.
    
    Returns:
        `log_prob_grad_fn`

    :param dict z: dictionary of parameter values keyed by site name.
    :return: tuple of `(z_grads, log_prob)`, where `z_grads` is a dictionary
        with the same keys as `z` containing gradients and log prob is a
        torch scalar.
    """

    def log_prob_grad_fn(z):
        z_keys, z_nodes = zip(*z.items())
        for node in z_nodes:
            node.requires_grad_(True)
        try:
            log_prob = log_prob_fn(z)

        # deal with singular matrices
        except RuntimeError as e:
            if "singular U" in str(e):
                grads = {k: v.new_zeros(v.shape) for k, v in z.items()}
                return grads, z_nodes[0].new_tensor(float("nan"))
            else:
                raise e

        grads = grad(log_prob, z_nodes)
        for node in z_nodes:
            node.requires_grad_(False)
        return dict(zip(z_keys, grads)), log_prob.detach()

    return log_prob_grad_fn


class TraceEinsumEvaluator:
    """
    Computes the log probability density of a trace (of a model with
    tree structure) that possibly contains discrete sample sites
    enumerated in parallel. This uses optimized `einsum` operations
    to marginalize out the the enumerated dimensions in the trace
    via :class:`~pyro.ops.contract.contract_to_tensor`.

    :param model_trace: execution trace from a static model.
    :param bool has_enumerable_sites: whether the trace contains any
        discrete enumerable sites.
    :param int max_plate_nesting: Optional bound on max number of nested
        :func:`pyro.plate` contexts.
    """

    def __init__(self, model_trace, has_enumerable_sites=False, max_plate_nesting=None):
        self.has_enumerable_sites = has_enumerable_sites
        self.max_plate_nesting = max_plate_nesting
        # To be populated using the model trace once.
        self._enum_dims = set()
        self.ordering = {}
        self._populate_cache(model_trace)

    def _populate_cache(self, model_trace):
        """
        Populate the ordinals (set of ``CondIndepStack`` frames)
        and enum_dims for each sample site.
        """
        if not self.has_enumerable_sites:
            return
        if self.max_plate_nesting is None:
            raise ValueError(
                "Finite value required for `max_plate_nesting` when model "
                "has discrete (enumerable) sites."
            )
        model_trace.compute_log_prob()
        model_trace.pack_tensors()
        for name, site in model_trace.nodes.items():
            if site["type"] == "sample" and not isinstance(site["fn"], _Subsample):
                if is_validation_enabled():
                    check_site_shape(site, self.max_plate_nesting)
                self.ordering[name] = frozenset(
                    model_trace.plate_to_symbol[f.name]
                    for f in site["cond_indep_stack"]
                    if f.vectorized
                )
        self._enum_dims = set(model_trace.symbol_to_dim) - set(
            model_trace.plate_to_symbol.values()
        )

    def _get_log_factors(self, model_trace):
        """
        Aggregates the `log_prob` terms into a list for each
        ordinal.
        """
        model_trace.compute_log_prob()
        model_trace.pack_tensors()
        log_probs = OrderedDict()
        # Collect log prob terms per independence context.
        for name, site in model_trace.nodes.items():
            if site["type"] == "sample" and not isinstance(site["fn"], _Subsample):
                if is_validation_enabled():
                    check_site_shape(site, self.max_plate_nesting)
                log_probs.setdefault(self.ordering[name], []).append(
                    site["packed"]["log_prob"]
                )
        return log_probs

    def log_prob(self, model_trace):
        """
        Returns the log pdf of `model_trace` by appropriately handling
        enumerated log prob factors.
        :return: log pdf of the trace.
        """
        if not self.has_enumerable_sites:
            return model_trace.log_prob_sum()
        log_probs = self._get_log_factors(model_trace)
        with shared_intermediates() as cache:
            return contract_to_tensor(log_probs, self._enum_dims, cache=cache)


def _guess_max_plate_nesting(model, args, kwargs):
    """
    Guesses max_plate_nesting by running the model once
    without enumeration. This optimistically assumes static model
    structure.
    """
    with poutine.block():
        model_trace = poutine.trace(model).get_trace(*args, **kwargs)
    sites = [site for site in model_trace.nodes.values() if site["type"] == "sample"]

    dims = [
        frame.dim
        for site in sites
        for frame in site["cond_indep_stack"]
        if frame.vectorized
    ]
    max_plate_nesting = -min(dims) if dims else 0
    return max_plate_nesting



from pathlib import Path
from typing import Iterable, Optional, Union

import numpy as np
import pandas as pd
import torch

from sbibm.utils.torch import get_default_device


def get_float_from_csv(
    path: Union[str, Path], dtype: type = np.float32,
):
    """Get a single float from a csv file
    """
    with open(path, "r") as fh:
        return np.loadtxt(fh).astype(dtype)


def get_results(
    dataset: str = "main_paper.csv", subfolder: str = "benchmarking_sbi/results/"
) -> pd.DataFrame:
    """Get results from https://github.com/sbi-benchmark/results/

    Args:
        dataset: Filename for dataset
        subfolder: Subfolder in repo

    Returns:
        Dataframe
    """
    df = pd.read_csv(
        f"https://raw.githubusercontent.com/sbi-benchmark/results/main/{subfolder}{dataset}"
    )
    return df


def get_tensor_from_csv(
    path: Union[str, Path], dtype: type = np.float32, atleast_2d: bool = True
) -> torch.Tensor:
    """Get `torch.Tensor` from csv at given path
    """
    device = get_default_device()

    if atleast_2d:
        return torch.from_numpy(np.atleast_2d(pd.read_csv(path)).astype(dtype)).to(
            device
        )
    else:
        return torch.from_numpy(pd.read_csv(path).astype(dtype)).to(device)


def get_ndarray_from_csv(
    path: Union[str, Path], dtype: type = np.float32, atleast_2d: bool = True
) -> np.ndarray:
    """Get `np.ndarray` from csv at given path
    """
    if atleast_2d:
        return np.atleast_2d(pd.read_csv(path)).astype(dtype)
    else:
        return pd.read_csv(path).astype(dtype)


def save_float_to_csv(
    path: Union[str, Path], data: float, dtype: type = np.float32,
):
    """Save a single float to a csv file
    """
    np.savetxt(
        path, np.asarray(data).reshape(-1).astype(np.float32), delimiter=",",
    )


def save_tensor_to_csv(
    path: Union[str, Path],
    data: torch.Tensor,
    columns: Optional[Iterable[str]] = None,
    dtype: type = np.float32,
    index: bool = False,
):
    """Save torch.Tensor to csv at given path
    """
    pd.DataFrame(data.cpu().numpy().astype(dtype), columns=columns,).to_csv(
        path, index=index
    )



from sbibm.algorithms.sbi.mcabc import run as mcabc
from sbibm.algorithms.sbi.snpe import run as snpe
from sbibm.algorithms.sbi.smcabc import run as smcabc
from sbibm.algorithms.sbi.snle import run as snle
from sbibm.algorithms.sbi.snre import run as snre

rej_abc = mcabc
smc_abc = smcabc



from typing import Optional, Tuple

import pandas as pd
import torch
from sbi.inference import SMCABC
from sklearn.linear_model import LinearRegression

import sbibm
from sbibm.tasks.task import Task
from sbibm.utils.kde import get_kde

from .utils import clip_int, get_sass_transform, run_lra


def run(
    task: Task,
    num_samples: int,
    num_simulations: int,
    num_observation: Optional[int] = None,
    observation: Optional[torch.Tensor] = None,
    population_size: Optional[int] = None,
    distance: str = "l2",
    epsilon_decay: float = 0.2,
    distance_based_decay: bool = True,
    ess_min: Optional[float] = None,
    initial_round_factor: int = 5,
    batch_size: int = 1000,
    kernel: str = "gaussian",
    kernel_variance_scale: float = 0.5,
    use_last_pop_samples: bool = True,
    algorithm_variant: str = "C",
    save_summary: bool = False,
    sass: bool = False,
    sass_fraction: float = 0.5,
    sass_feature_expansion_degree: int = 3,
    lra: bool = False,
    lra_sample_weights: bool = True,
    kde_bandwidth: Optional[str] = "cv",
    kde_sample_weights: bool = False,
) -> Tuple[torch.Tensor, int, Optional[torch.Tensor]]:
    """Runs SMC-ABC from `sbi`

    SMC-ABC supports two different ways of scheduling epsilon:
    1) Exponential decay: eps_t+1 = epsilon_decay * eps_t
    2) Distance based decay: the new eps is determined from the "epsilon_decay" 
        quantile of the distances of the accepted simulations in the previous population. This is used if `distance_based_decay` is set to True.

    Args:
        task: Task instance
        num_samples: Number of samples to generate from posterior
        num_simulations: Simulation budget
        num_observation: Observation number to load, alternative to `observation`
        observation: Observation, alternative to `num_observation`
        population_size: If None, uses heuristic: 1000 if `num_simulations` is greater
            than 10k, else 100
        distance: Distance function, options = {l1, l2, mse}
        epsilon_decay: Decay for epsilon; treated as quantile in case of distance based decay.
        distance_based_decay: Whether to determine new epsilon from quantile of
            distances of the previous population.
        ess_min: Threshold for resampling a population if effective sampling size is 
            too small.
        initial_round_factor: Used to determine initial round size
        batch_size: Batch size for the simulator
        kernel: Kernel distribution used to perturb the particles.
        kernel_variance_scale: Scaling factor for kernel variance.
        use_last_pop_samples: If True, samples of a population that was quit due to
            budget are used by filling up missing particles from the previous
            population.
        algorithm_variant: There are three SMCABC variants implemented: A, B, and C.
            See doctstrings in SBI package for more details.
        save_summary: Whether to save a summary containing all populations, distances,
            etc. to file.
        sass: If True, summary statistics are learned as in
            Fearnhead & Prangle 2012.
        sass_fraction: Fraction of simulation budget to use for sass.
        sass_feature_expansion_degree: Degree of polynomial expansion of the summary
            statistics.
        lra: If True, posterior samples are adjusted with
            linear regression as in Beaumont et al. 2002.
        lra_sample_weights: Whether to weigh LRA samples
        kde_bandwidth: If not None, will resample using KDE when necessary, set
            e.g. to "cv" for cross-validated bandwidth selection
        kde_sample_weights: Whether to weigh KDE samples


    Returns:
        Samples from posterior, number of simulator calls, log probability of true params if computable
    """
    assert not (num_observation is None and observation is None)
    assert not (num_observation is not None and observation is not None)

    log = sbibm.get_logger(__name__)
    smc_papers = dict(A="Toni 2010", B="Sisson et al. 2007", C="Beaumont et al. 2009")
    log.info(f"Running SMC-ABC as in {smc_papers[algorithm_variant]}.")

    prior = task.get_prior_dist()
    simulator = task.get_simulator(max_calls=num_simulations)
    if observation is None:
        observation = task.get_observation(num_observation)

    if population_size is None:
        population_size = 100
        if num_simulations > 10_000:
            population_size = 1000

    population_size = min(population_size, num_simulations)

    initial_round_size = clip_int(
        value=initial_round_factor * population_size,
        minimum=population_size,
        maximum=max(0.5 * num_simulations, population_size),
    )

    inference_method = SMCABC(
        simulator=simulator,
        prior=prior,
        simulation_batch_size=batch_size,
        distance=distance,
        show_progress_bars=True,
        kernel=kernel,
        algorithm_variant=algorithm_variant,
    )
    posterior, summary = inference_method(
        x_o=observation,
        num_particles=population_size,
        num_initial_pop=initial_round_size,
        num_simulations=num_simulations,
        epsilon_decay=epsilon_decay,
        distance_based_decay=distance_based_decay,
        ess_min=ess_min,
        kernel_variance_scale=kernel_variance_scale,
        use_last_pop_samples=use_last_pop_samples,
        return_summary=True,
        lra=lra,
        lra_with_weights=lra_sample_weights,
        sass=sass,
        sass_fraction=sass_fraction,
        sass_expansion_degree=sass_feature_expansion_degree,
    )

    if save_summary:
        log.info("Saving smcabc summary to csv.")
        pd.DataFrame.from_dict(summary,).to_csv("summary.csv", index=False)

    assert simulator.num_simulations == num_simulations

    if kde_bandwidth is not None:
        samples = posterior._samples

        log.info(
            f"KDE on {samples.shape[0]} samples with bandwidth option {kde_bandwidth}"
        )

        kde = get_kde(
            samples,
            bandwidth=kde_bandwidth,
            sample_weight=posterior._log_weights.exp() if kde_sample_weights else None,
        )
        samples = kde.sample(num_samples)
    else:
        samples = posterior.sample((num_samples,)).detach()

    if num_observation is not None:
        true_parameters = task.get_true_parameters(num_observation=num_observation)
        log_prob_true_parameters = posterior.log_prob(true_parameters)
        return samples, simulator.num_simulations, log_prob_true_parameters
    else:
        return samples, simulator.num_simulations, None



"""A global module containing functions for managing the project."""

from future import standard_library
standard_library.install_aliases()
__author__ = 'wittawat'

import kgof
import os
import pickle


def get_root():
    """Return the full path to the root of the package"""
    return os.path.abspath(os.path.dirname(kgof.__file__))

def result_folder():
    """Return the full path to the result/ folder containing experimental result 
    files"""
    import kgof.config as config
    results_path = config.expr_configs['expr_results_path']
    return results_path
    #return os.path.join(get_root(), 'result')

def data_folder():
    """
    Return the full path to the data folder 
    """
    import kgof.config as config
    data_path = config.expr_configs['data_path']
    return data_path
    #return os.path.join(get_root(), 'data')

def data_file(*relative_path):
    """
    Access the file under the data folder. The path is relative to the 
    data folder
    """
    dfolder = data_folder()
    return os.path.join(dfolder, *relative_path)

def load_data_file(*relative_path):
    fpath = data_file(*relative_path)
    return pickle_load(fpath)

def ex_result_folder(ex):
    """Return the full path to the folder containing result files of the specified 
    experiment. 
    ex: a positive integer. """
    rp = result_folder()
    fpath = os.path.join(rp, 'ex%d'%ex )
    if not os.path.exists(fpath):
        create_dirs(fpath)
    return fpath

def create_dirs(full_path):
    """Recursively create the directories along the specified path. 
    Assume that the path refers to a folder. """
    if not os.path.exists(full_path):
        os.makedirs(full_path)

def ex_result_file(ex, *relative_path ):
    """Return the full path to the file identified by the relative path as a list 
    of folders/files under the result folder of the experiment ex. """
    rf = ex_result_folder(ex)
    return os.path.join(rf, *relative_path)

def ex_save_result(ex, result, *relative_path):
    """Save a dictionary object result for the experiment ex. Serialization is 
    done with pickle. 
    EX: ex_save_result(1, result, 'data', 'result.p'). Save under result/ex1/data/result.p 
    EX: ex_save_result(1, result, 'result.p'). Save under result/ex1/result.p 
    """
    fpath = ex_result_file(ex, *relative_path)
    dir_path = os.path.dirname(fpath)
    create_dirs(dir_path)
    # 
    with open(fpath, 'wb') as f:
        # expect result to be a dictionary
        pickle.dump(result, f)

def ex_load_result(ex, *relative_path):
    """Load a result identified by the  path from the experiment ex"""
    fpath = ex_result_file(ex, *relative_path)
    return pickle_load(fpath)

def ex_file_exists(ex, *relative_path):
    """Return true if the result file in under the specified experiment folder
    exists"""
    fpath = ex_result_file(ex, *relative_path)
    return os.path.isfile(fpath)

def pickle_load(fpath):
    if not os.path.isfile(fpath):
        raise ValueError('%s does not exist' % fpath)

    with open(fpath, 'rb') as f:
        # expect a dictionary
        result = pickle.load(f)
    return result




from pathlib import Path
from typing import Callable, Dict, List, Optional

import numpy as np
import pyro
import pyro.distributions as pdist
import torch

from sbibm.tasks.simulator import Simulator
from sbibm.tasks.task import Task
from sbibm.utils.io import get_tensor_from_csv
from sbibm.utils.torch import get_default_device


class BernoulliGLM(Task):
    def __init__(self, summary="sufficient"):
        """Bernoulli GLM
        """
        self.summary = summary
        if self.summary == "sufficient":
            dim_data = 10
            name_display = "Bernoulli GLM"
            self.raw = False
        elif self.summary == "raw":
            dim_data = 100
            self.raw = True
            name_display = "Bernoulli GLM Raw"
        else:
            raise NotImplementedError

        super().__init__(
            dim_parameters=10,
            dim_data=dim_data,
            name=Path(__file__).parent.name,
            name_display=name_display,
            num_simulations=[1000, 10000, 100000, 1000000],
            num_posterior_samples=10000,
            num_observations=10,
            path=Path(__file__).parent.absolute(),
        )

        self.stimulus = {
            "dt": 1,  # timestep
            "duration": 100,  # duration of input stimulus
            "seed": 42,  # seperate seed to freeze noise on input current
        }

        # Prior on offset and filter
        # Smoothness in filter encouraged by penalyzing 2nd order differences
        M = self.dim_parameters - 1
        D = torch.diag(torch.ones(M)) - torch.diag(torch.ones(M - 1), -1)
        F = torch.matmul(D, D) + torch.diag(1.0 * torch.arange(M) / (M)) ** 0.5
        Binv = torch.zeros(size=(M + 1, M + 1))
        Binv[0, 0] = 0.5  # offset
        Binv[1:, 1:] = torch.matmul(F.T, F)  # filter

        self.prior_params = {"loc": torch.zeros((M + 1,)), "precision_matrix": Binv}
        self.prior_dist = pdist.MultivariateNormal(**self.prior_params)

    def get_prior(self) -> Callable:
        def prior(num_samples=1):
            return pyro.sample("parameters", self.prior_dist.expand_by([num_samples]))

        return prior

    def get_simulator(self, max_calls: Optional[int] = None) -> Simulator:
        """Get function returning samples from simulator given parameters

        Args:
            max_calls: Maximum number of function calls. Additional calls will 
                result in SimulationBudgetExceeded exceptions. Defaults to None 
                for infinite budget

        Return:
            Simulator callable
        """
        device = get_default_device()

        stimulus_I = torch.load(self.path / "files" / "stimulus_I.pt").to(device)
        design_matrix = torch.load(self.path / "files" / "design_matrix.pt").to(device)

        def simulator(
            parameters: torch.Tensor, return_both: bool = False
        ) -> torch.Tensor:
            """Simulates model for given parameters

            If `return_both` is True, will additionally return spike train not reduced to summary features
            """

            data = []
            data_raw = []
            for b in range(parameters.shape[0]):
                # Simulate GLM
                psi = torch.matmul(design_matrix, parameters[b, :])
                z = 1 / (1 + torch.exp(-psi))
                y = (torch.rand(design_matrix.shape[0]) < z).float()

                # Calculate summary statistics
                num_spikes = torch.sum(y).unsqueeze(0)
                sta = torch.nn.functional.conv1d(
                    y.reshape(1, 1, -1), stimulus_I.reshape(1, 1, -1), padding=8
                ).squeeze()[-9:]
                data.append(torch.cat((num_spikes, sta)))

                if self.raw or return_both:
                    data_raw.append(y)

            if not return_both:
                if not self.raw:
                    return torch.stack(data)
                else:
                    return torch.stack(data_raw)
            else:
                return torch.stack(data), torch.stack(data_raw)

        return Simulator(task=self, simulator=simulator, max_calls=max_calls)

    def get_observation(self, num_observation: int) -> torch.Tensor:
        """Get observed data for a given observation number
        """
        if not self.raw:
            path = (
                self.path
                / "files"
                / f"num_observation_{num_observation}"
                / "observation.csv"
            )
            return get_tensor_from_csv(path)
        else:
            path = (
                self.path
                / "files"
                / f"num_observation_{num_observation}"
                / "observation_raw.csv"
            )
            return get_tensor_from_csv(path)

    def flatten_data(self, data: torch.Tensor) -> torch.Tensor:
        """Flattens data

        Data returned by the simulator is always flattened into 2D Tensors
        """
        if type(data) == tuple:
            return data
        else:
            return data.reshape(-1, self.dim_data)

    def _sample_reference_posterior(
        self, num_samples: int, num_observation: Optional[int] = None,
    ) -> torch.Tensor:
        from pypolyagamma import PyPolyaGamma
        from tqdm import tqdm

        self.dim_data = 10
        # stimulus_I = torch.load(self.path / "files" / "stimulus_I.pt")
        design_matrix = torch.load(self.path / "files" / "design_matrix.pt")
        true_parameters = self.get_true_parameters(num_observation)
        self.raw = True
        observation_raw = self.get_observation(num_observation)
        self.raw = False

        mcmc_num_samples_warmup = 25000
        mcmc_thinning = 25
        mcmc_num_samples = mcmc_num_samples_warmup + mcmc_thinning * num_samples

        pg = PyPolyaGamma()
        X = design_matrix.numpy()
        obs = observation_raw.numpy()
        Binv = self.prior_params["precision_matrix"].numpy()

        sample = true_parameters.numpy().reshape(-1)  # Init at true parameters
        samples = []
        for j in tqdm(range(mcmc_num_samples)):
            psi = np.dot(X, sample)
            w = np.array([pg.pgdraw(1, b) for b in psi])
            O = np.diag(w)  # noqa: E741
            V = np.linalg.inv(np.dot(np.dot(X.T, O), X) + Binv)
            m = np.dot(V, np.dot(X.T, obs.reshape(-1) - 1 * 0.5))
            sample = np.random.multivariate_normal(np.ravel(m), V)
            samples.append(sample)
        samples = np.asarray(samples).astype(np.float32)
        samples_subset = samples[mcmc_num_samples_warmup::mcmc_thinning, :]

        reference_posterior_samples = torch.from_numpy(samples_subset)

        return reference_posterior_samples

    def _setup(self, regenerate_stimulus=False):
        """Setup the task: generate observations and reference posterior samples

        In most cases, you don't need to execute this method, since its results are stored to disk.
        Re-executing will overwrite existing files.

        Reference samples are constructed using Polya-Gamma MCMC. The sampler consists of two iterative Gibbs updates:
        1. sample auxiliary variables: w ~ PG(N, psi)
        2. sample parameters: beta ~ N(m, V); V = inv(X'O X + Binv), m = V*(X'k), k = y - N/2

        Note that running this method requires pypolyagamma, see https://github.com/slinderman/pypolyagamma
        for installation instructions.

        There is an open issue leading to errors on pip install, see:
        https://github.com/slinderman/pypolyagamma/issues/36

        Manual installation with the following steps succeeded:
        pip install cython==0.28
        git clone git@github.com:slinderman/pypolyagamma.git
        cd pypolyagamma
        pip install -e .

        Takes about 1-2 minutes on a laptop per observation
        """
        # Generate input stimulus (same across all observations)
        # Stimulus is Gaussian white noise ~N(0, 1)
        if regenerate_stimulus:
            stimulus_t = torch.arange(
                0, self.stimulus["duration"], self.stimulus["dt"], dtype=torch.float32
            )
            path = self.path / "files" / "stimulus_t.pt"
            path.parent.mkdir(parents=True, exist_ok=True)
            torch.save(stimulus_t, path)
            stimulus_I = torch.from_numpy(
                np.random.RandomState(self.stimulus["seed"])
                .randn(len(stimulus_t))
                .reshape(-1)
                .astype(np.float32)
            )
            path = self.path / "files" / "stimulus_I.pt"
            path.parent.mkdir(parents=True, exist_ok=True)
            torch.save(stimulus_I, path)

            # Build design matrix X, such that X * h returns convolution of x with filter h
            # Including linear offset by first element
            design_matrix = torch.zeros(size=(len(stimulus_t), self.dim_parameters - 1))
            for j in range(self.dim_parameters - 1):
                design_matrix[j:, j] = stimulus_I[0 : len(stimulus_t) - j]
            design_matrix = torch.cat(
                (torch.ones(size=(len(stimulus_t), 1)), design_matrix), axis=1
            )
            path = self.path / "files" / "design_matrix.pt"
            path.parent.mkdir(parents=True, exist_ok=True)
            torch.save(design_matrix, path)

        observation_seeds = np.arange(self.num_observations, dtype=np.int) + int(1e6)
        for num_observation, observation_seed in enumerate(observation_seeds, start=1):
            np.random.seed(observation_seed)
            torch.manual_seed(observation_seed)
            self._save_observation_seed(num_observation, observation_seed)

            prior = self.get_prior()
            true_parameters = prior(num_samples=1)
            self._save_true_parameters(num_observation, true_parameters)

            simulator = self.get_simulator()
            observation, observation_raw = simulator(true_parameters, return_both=True)
            self._save_observation(num_observation, observation)

            path = (
                self.path
                / "files"
                / f"num_observation_{num_observation}"
                / "observation_raw.csv"
            )
            self.dim_data = 100
            self.save_data(path, observation_raw)
            self.dim_data = 10

            reference_posterior_samples = self._sample_reference_posterior(
                num_samples=10_000, num_observation=num_observation
            )

            self._save_reference_posterior_samples(
                num_observation, reference_posterior_samples
            )


if __name__ == "__main__":
    task = BernoulliGLM()
    task._setup()



"""Module containing convenient functions for plotting"""

from builtins import range
from builtins import object
__author__ = 'wittawat'

import kgof.glo as glo
import matplotlib
import matplotlib.pyplot as plt
import autograd.numpy as np


def get_func_tuples():
    """
    Return a list of tuples where each tuple is of the form
        (func_name used in the experiments, label name, plot line style)
    """
    func_tuples = [
            ('job_fssdJ1q_med', 'FSSD-rand J1', 'r--^'),
            ('job_fssdJ5q_med', 'FSSD-rand', 'r--^'),
            ('job_fssdq_med', 'FSSD-rand', 'r--^'),

            ('job_fssdJ1q_opt', 'FSSD-opt J1', 'r-s'),
            ('job_fssdq_opt', 'FSSD-opt', 'r-s'),
            ('job_fssdJ5q_opt', 'FSSD-opt', 'r-s'),
            ('job_fssdJ5q_imq_optv', 'FSSD-IMQv', 'k-h'),
            ('job_fssdJ5q_imqb1_optv', 'FSSD-IMQ-1', 'k--s'),
            ('job_fssdJ5q_imqb2_optv', 'FSSD-IMQ-2', 'k-->'),
            ('job_fssdJ5q_imqb3_optv', 'FSSD-IMQ-3', 'k-.*'),
            ('job_fssdJ5q_imq_opt', 'FSSD-IMQ', 'y-x'),
            ('job_fssdJ5q_imq_optbv', 'FSSD-IMQ-bv', 'y--d'),
            ('job_fssdJ10q_opt', 'FSSD-opt', 'k-s'),

            ('job_fssdJ5p_opt', 'FSSD-opt J5', 'm-s'),
            ('job_fssdp_opt', 'FSSDp-opt', 'm-s'),
            ('job_fssdJ10p_opt', 'FSSDp-opt J10', 'k-s'),

            ('job_fssdJ1q_opt2', 'FSSD-opt2 J1', 'b-^'),
            ('job_fssdJ5q_opt2', 'FSSD-opt2 J5', 'r-^'),
            ('job_me_opt', 'ME-opt', 'b-d'),

            ('job_kstein_med', 'KSD', 'g-o'),
            ('job_kstein_imq', 'KSD-IMQ', 'c-*'),
            ('job_lin_kstein_med', 'LKS', 'g-.h'),
            ('job_mmd_med', 'MMD', 'm--^'),
            ('job_mmd_opt', 'MMD-opt', 'm-<'),
            ('job_mmd_dgauss_opt', 'MMD-dopt', 'y-<'),
            ]
    return func_tuples

def set_default_matplotlib_options():
    # font options
    font = {
    #     'family' : 'normal',
        #'weight' : 'bold',
        'size'   : 30
    }
    matplotlib.rc('font', **{'family': 'serif', 'serif': ['Computer Modern']})


    # matplotlib.use('cairo')
    matplotlib.rc('text', usetex=True)
    matplotlib.rcParams['text.usetex'] = True
    plt.rc('font', **font)
    plt.rc('lines', linewidth=3, markersize=10)
    # matplotlib.rcParams['ps.useafm'] = True
    # matplotlib.rcParams['pdf.use14corefonts'] = True

    matplotlib.rcParams['pdf.fonttype'] = 42
    matplotlib.rcParams['ps.fonttype'] = 42

def get_func2label_map():
    # map: job_func_name |-> plot label
    func_tuples = get_func_tuples()
    #M = {k:v for (k,v) in zip(func_names, labels)}
    M = {k:v for (k,v,_) in func_tuples}
    return M


def func_plot_fmt_map():
    """
    Return a map from job function names to matplotlib plot styles 
    """
    # line_styles = ['o-', 'x-',  '*-', '-_', 'D-', 'h-', '+-', 's-', 'v-', 
    #               ',-', '1-']
    func_tuples = get_func_tuples()
    M = {k:v for (k, _, v) in func_tuples}
    return M


class PlotValues(object):
    """
    An object encapsulating values of a plot where there are many curves, 
    each corresponding to one method, with common x-axis values.
    """
    def __init__(self, xvalues, methods, plot_matrix):
        """
        xvalues: 1d numpy array of x-axis values
        methods: a list of method names
        plot_matrix: len(methods) x len(xvalues) 2d numpy array containing 
            values that can be used to plot
        """
        self.xvalues = xvalues
        self.methods = methods
        self.plot_matrix = plot_matrix

    def ascii_table(self, tablefmt="pipe"):
        """
        Return an ASCII string representation of the table.

        tablefmt: "plain", "fancy_grid", "grid", "simple" might be useful.
        """
        methods = self.methods
        xvalues = self.xvalues
        plot_matrix = self.plot_matrix

        import tabulate
        # https://pypi.python.org/pypi/tabulate
        aug_table = np.hstack((np.array(methods)[:, np.newaxis], plot_matrix))
        return tabulate.tabulate(aug_table, xvalues, tablefmt=tablefmt)

# end of class PlotValues

def plot_prob_reject(ex, fname, func_xvalues, xlabel, func_title=None, 
        return_plot_values=False):
    """
    plot the empirical probability that the statistic is above the threshold.
    This can be interpreted as type-1 error (when H0 is true) or test power 
    (when H1 is true). The plot is against the specified x-axis.

    - ex: experiment number 
    - fname: file name of the aggregated result
    - func_xvalues: function taking aggregated results dictionary and return the values 
        to be used for the x-axis values.            
    - xlabel: label of the x-axis. 
    - func_title: a function: results dictionary -> title of the plot
    - return_plot_values: if true, also return a PlotValues as the second
      output value.

    Return loaded results
    """
    #from IPython.core.debugger import Tracer 
    #Tracer()()

    results = glo.ex_load_result(ex, fname)

    def rej_accessor(jr):
        rej = jr['test_result']['h0_rejected']
        # When used with vectorize(), making the value float will make the resulting 
        # numpy array to be of float. nan values can be stored.
        return float(rej)

    #value_accessor = lambda job_results: job_results['test_result']['h0_rejected']
    vf_pval = np.vectorize(rej_accessor)
    # results['job_results'] is a dictionary: 
    # {'test_result': (dict from running perform_test(te) '...':..., }
    rejs = vf_pval(results['job_results'])
    repeats, _, n_methods = results['job_results'].shape

    # yvalues (corresponding to xvalues) x #methods
    mean_rejs = np.mean(rejs, axis=0)
    #print mean_rejs
    #std_pvals = np.std(rejs, axis=0)
    #std_pvals = np.sqrt(mean_rejs*(1.0-mean_rejs))

    xvalues = func_xvalues(results)

    #ns = np.array(results[xkey])
    #te_proportion = 1.0 - results['tr_proportion']
    #test_sizes = ns*te_proportion
    line_styles = func_plot_fmt_map()
    method_labels = get_func2label_map()
    
    func_names = [f.__name__ for f in results['method_job_funcs'] ]
    plotted_methods = []
    for i in range(n_methods):    
        te_proportion = 1.0 - results['tr_proportion']
        fmt = line_styles[func_names[i]]
        #plt.errorbar(ns*te_proportion, mean_rejs[:, i], std_pvals[:, i])
        method_label = method_labels[func_names[i]]
        plotted_methods.append(method_label)
        plt.plot(xvalues, mean_rejs[:, i], fmt, label=method_label)
    '''
    else:
        # h0 is true 
        z = stats.norm.isf( (1-confidence)/2.0)
        for i in range(n_methods):
            phat = mean_rejs[:, i]
            conf_iv = z*(phat*(1-phat)/repeats)**0.5
            #plt.errorbar(test_sizes, phat, conf_iv, fmt=line_styles[i], label=method_labels[i])
            plt.plot(test_sizes, mean_rejs[:, i], line_styles[i], label=method_labels[i])
    '''
            
    ylabel = 'Rejection rate'
    plt.ylabel(ylabel)
    plt.xlabel(xlabel)
    plt.xticks(np.hstack((xvalues) ))
    
    alpha = results['alpha']
    plt.legend(loc='best')
    title = '%s. %d trials. $\\alpha$ = %.2g.'%( results['prob_label'],
            repeats, alpha) if func_title is None else func_title(results)
    plt.title(title)
    plt.grid()
    if return_plot_values:
        return results, PlotValues(xvalues=xvalues, methods=plotted_methods,
                plot_matrix=mean_rejs.T)
    else:
        return results
        

def plot_runtime(ex, fname, func_xvalues, xlabel, func_title=None):
    results = glo.ex_load_result(ex, fname)
    value_accessor = lambda job_results: job_results['time_secs']
    vf_pval = np.vectorize(value_accessor)
    # results['job_results'] is a dictionary: 
    # {'test_result': (dict from running perform_test(te) '...':..., }
    times = vf_pval(results['job_results'])
    repeats, _, n_methods = results['job_results'].shape
    time_avg = np.mean(times, axis=0)
    time_std = np.std(times, axis=0)

    xvalues = func_xvalues(results)

    #ns = np.array(results[xkey])
    #te_proportion = 1.0 - results['tr_proportion']
    #test_sizes = ns*te_proportion
    line_styles = func_plot_fmt_map()
    method_labels = get_func2label_map()
    
    func_names = [f.__name__ for f in results['method_job_funcs'] ]
    for i in range(n_methods):    
        te_proportion = 1.0 - results['tr_proportion']
        fmt = line_styles[func_names[i]]
        #plt.errorbar(ns*te_proportion, mean_rejs[:, i], std_pvals[:, i])
        method_label = method_labels[func_names[i]]
        plt.errorbar(xvalues, time_avg[:, i], yerr=time_std[:,i], fmt=fmt,
                label=method_label)
            
    ylabel = 'Time (s)'
    plt.ylabel(ylabel)
    plt.xlabel(xlabel)
    plt.xlim([np.min(xvalues), np.max(xvalues)])
    plt.xticks( xvalues, xvalues )
    plt.legend(loc='best')
    plt.gca().set_yscale('log')
    title = '%s. %d trials. '%( results['prob_label'],
            repeats ) if func_title is None else func_title(results)
    plt.title(title)
    #plt.grid()
    return results


def box_meshgrid(func, xbound, ybound, nx=50, ny=50):
    """
    Form a meshed grid (to be used with a contour plot) on a box
    specified by xbound, ybound. Evaluate the grid with [func]: (n x 2) -> n.
    
    - xbound: a tuple (xmin, xmax)
    - ybound: a tuple (ymin, ymax)
    - nx: number of points to evluate in the x direction
    
    return XX, YY, ZZ where XX is a 2D nd-array of size nx x ny
    """
    
    # form a test location grid to try 
    minx, maxx = xbound
    miny, maxy = ybound
    loc0_cands = np.linspace(minx, maxx, nx)
    loc1_cands = np.linspace(miny, maxy, ny)
    lloc0, lloc1 = np.meshgrid(loc0_cands, loc1_cands)
    # nd1 x nd0 x 2
    loc3d = np.dstack((lloc0, lloc1))
    # #candidates x 2
    all_loc2s = np.reshape(loc3d, (-1, 2) )
    # evaluate the function
    func_grid = func(all_loc2s)
    func_grid = np.reshape(func_grid, (ny, nx))
    
    assert lloc0.shape[0] == ny
    assert lloc0.shape[1] == nx
    assert np.all(lloc0.shape == lloc1.shape)
    
    return lloc0, lloc1, func_grid

def get_density_cmap():
    """
    Return a colormap for plotting the model density p.
    Red = high density 
    white = very low density.
    Varying from white (low) to red (high).
    """
    # Add completely white color to Reds colormap in Matplotlib
    list_colors = plt.cm.datad['Reds']
    list_colors = list(list_colors)
    list_colors.insert(0, (1, 1, 1))
    list_colors.insert(0, (1, 1, 1))
    lscm = matplotlib.colors.LinearSegmentedColormap.from_list("my_Reds", list_colors)
    return lscm



import logging
import math
from typing import Callable, Optional

import numpy as np
import torch
from sbibm.third_party.kgof.kernel import KGauss
from sbibm.third_party.kgof.goftest import KernelSteinTest, bootstrapper_rademacher
from sbibm.third_party.kgof.util import meddistance

from sbibm.tasks.task import Task
from sbibm.utils.torch import get_default_device

log = logging.getLogger(__name__)


def ksd(
    task: Task,
    num_observation: int,
    samples: torch.Tensor,
    sig2: Optional[float] = None,
    log: bool = True,
) -> torch.Tensor:
    """Gets `log_prob_grad_fn` from task and runs KSD

    Args:
        task: Task
        num_observation: Observation
        samples: Samples
        sig2: Length scale
        log: Whether to log test result

    Returns:
        The test result is returned
    """
    try:
        device = get_default_device()
        site_name = "parameters"

        log_prob_grad_fn = task._get_log_prob_grad_fn(
            num_observation=num_observation,
            implementation="pyro",
            posterior=True,
            jit_compile=False,
            automatic_transform_enabled=True,
        )

        samples_transformed = task._get_transforms(automatic_transform_enabled=True)[
            site_name
        ](samples)

        def log_prob_grad_numpy(parameters: np.ndarray) -> np.ndarray:
            lpg = log_prob_grad_fn(
                torch.from_numpy(parameters.astype(np.float32)).to(device)
            )
            return lpg.detach().cpu().numpy()

        test_statistic = ksd_gaussian_kernel(
            log_prob_grad_numpy, samples_transformed, sig2=sig2
        )
        if log:
            return math.log(test_statistic)
        else:
            return test_statistic

    except:

        return torch.tensor(float("nan"))


def ksd_gaussian_kernel(
    log_prob_grad_fn: Callable,
    samples: torch.Tensor,
    sig2: Optional[float] = None,
    seed: int = 101,
) -> torch.Tensor:
    """KSD test with `kgof` package

    Args:
        log_prob_grad_fn: Function returning the gradient of the log probability.
            It receives torch.Tensors as inputs and should output a torch.Tensor
            as well.
        samples: Samples for the test as a torch.Tensor

    Returns:
        The test result is returned
    """
    density = UnnormalizedDensityWrapped(log_prob_grad=log_prob_grad_fn)
    samples = DataWrapped(samples)

    if sig2 is None:
        sig2 = meddistance(samples.data(), subsample=1000) ** 2
    else:
        sig2 = float(sig2)

    kernel = KGauss(sig2)

    kstein = KernelSteinTest(
        density,
        kernel,
        bootstrapper=bootstrapper_rademacher,
        alpha=0.01,
        n_simulate=500,
        seed=seed + 1,
    )
    test_result = kstein.perform_test(
        samples, return_simulated_stats=False, return_ustat_gram=False
    )
    # log.info(f"H0 rejected: {test_result['h0_rejected']}")

    # TODO: normalize by sample size?
    return torch.tensor(test_result["test_stat"])


class DataWrapped:
    def __init__(self, data):
        """Wraps `data` such that it can be used with `kgof`
        """
        self._data = data

    def data(self):
        return self._data.cpu().numpy()


class UnnormalizedDensityWrapped:
    def __init__(self, log_prob_grad: Callable):
        """Wraps `log_prob_grad` function such that it can be used with `kgof`
        """
        self.log_prob_grad = log_prob_grad

    def grad_log(self, X: np.ndarray) -> np.ndarray:
        return self.log_prob_grad(X)



import logging
import time
from math import ceil
from typing import Optional

import elfi
import numpy as np
import torch

from sbibm.algorithms.elfi.utils.prior import build_prior
from sbibm.algorithms.elfi.utils.wrapper import Simulator
from sbibm.tasks.task import Task


def run(
    task: Task,
    num_samples: int,
    num_simulations: int,
    num_observation: Optional[int] = None,
    observation: Optional[torch.Tensor] = None,
    num_chains: int = 10,
    num_warmup: int = 1000,
) -> (torch.Tensor, int, Optional[torch.Tensor]):
    """Runs BOLFI from elfi package
     
    Args:
        task: Task instance
        num_samples: Number of samples to generate from posterior
        num_simulations: Simulation budget
        num_observation: Observation number to load, alternative to `observation`
        observation: Observation, alternative to `num_observation`
        num_chains: Number of chains
        num_warmup: Warmup steps

    Returns:
        Samples from posterior, number of simulator calls, log probability of true params if computable
    """
    assert not (num_observation is None and observation is None)
    assert not (num_observation is not None and observation is not None)

    logging.basicConfig(level=logging.INFO)

    log = logging.getLogger(__name__)
    log.warn("ELFI is not fully supported yet!")

    # Initialize model object
    m = elfi.ElfiModel()

    # Prior
    bounds = build_prior(task=task, model=m)

    # Observation
    if observation is None:
        observation = task.get_observation(num_observation)
    observation = observation.numpy()

    # Simulator
    simulator = task.get_simulator(max_calls=num_simulations)
    elfi.Simulator(
        Simulator(simulator),
        *[m[f"parameter_{dim}"] for dim in range(task.dim_parameters)],
        observed=observation,
        name=task.name,
    )

    # Euclidean distance
    elfi.Distance("euclidean", m[task.name], name="distance")

    # Log distance
    elfi.Operation(np.log, m["distance"], name="log_distance")

    # Inference
    num_samples_per_chain = ceil(num_samples / num_chains)
    tic = time.time()
    bolfi = elfi.BOLFI(model=m, target_name="log_distance", bounds=bounds)
    bolfi.fit(n_evidence=num_simulations)
    result_BOLFI = bolfi.sample(
        num_samples_per_chain + num_warmup,
        warmup=num_warmup,
        n_chains=num_chains,
        info_freq=int(100),
    )
    toc = time.time()

    samples = torch.from_numpy(result_BOLFI.samples_array.astype(np.float32)).reshape(
        -1, task.dim_parameters
    )[:num_samples, :]

    assert samples.shape[0] == num_samples

    # TODO: return log prob of true parameters

    return samples, simulator.num_simulations, None



from typing import Optional

import numpy as np
import torch
from sklearn.model_selection import KFold, cross_val_score
from sklearn.neural_network import MLPClassifier


def c2st(
    X: torch.Tensor,
    Y: torch.Tensor,
    seed: int = 1,
    n_folds: int = 5,
    scoring: str = "accuracy",
    z_score: bool = True,
    noise_scale: Optional[float] = None,
) -> torch.Tensor:
    """Classifier-based 2-sample test returning accuracy

    Trains classifiers with N-fold cross-validation [1]. Scikit learn MLPClassifier are
    used, with 2 hidden layers of 10x dim each, where dim is the dimensionality of the
    samples X and Y.

    Args:
        X: Sample 1
        Y: Sample 2
        seed: Seed for sklearn
        n_folds: Number of folds
        z_score: Z-scoring using X
        noise_scale: If passed, will add Gaussian noise with std noise_scale to samples
    
    References:
        [1]: https://scikit-learn.org/stable/modules/cross_validation.html
    """
    if z_score:
        X_mean = torch.mean(X, axis=0)
        X_std = torch.std(X, axis=0)
        X = (X - X_mean) / X_std
        Y = (Y - X_mean) / X_std

    if noise_scale is not None:
        X += noise_scale * torch.randn(X.shape)
        Y += noise_scale * torch.randn(Y.shape)

    X = X.cpu().numpy()
    Y = Y.cpu().numpy()

    ndim = X.shape[1]

    clf = MLPClassifier(
        activation="relu",
        hidden_layer_sizes=(10 * ndim, 10 * ndim),
        max_iter=10000,
        solver="adam",
        random_state=seed,
    )

    data = np.concatenate((X, Y))
    target = np.concatenate((np.zeros((X.shape[0],)), np.ones((Y.shape[0],)),))

    shuffle = KFold(n_splits=n_folds, shuffle=True, random_state=seed)
    scores = cross_val_score(clf, data, target, cv=shuffle, scoring=scoring)

    scores = np.asarray(np.mean(scores)).astype(np.float32)
    return torch.from_numpy(np.atleast_1d(scores))


def c2st_auc(
    X: torch.Tensor,
    Y: torch.Tensor,
    seed: int = 1,
    n_folds: int = 5,
    z_score: bool = True,
    noise_scale: Optional[float] = None,
) -> torch.Tensor:
    """Classifier-based 2-sample test returning AUC (area under curve)

    Same as c2st, except that it returns ROC AUC rather than accuracy

    Args:
        X: Sample 1
        Y: Sample 2
        seed: Seed for sklearn
        n_folds: Number of folds
        z_score: Z-scoring using X
        noise_scale: If passed, will add Gaussian noise with std noise_scale to samples
    
    Returns:
        Metric
    """
    return c2st(
        X,
        Y,
        seed=seed,
        n_folds=n_folds,
        scoring="roc_auc",
        z_score=z_score,
        noise_scale=noise_scale,
    )



"""
Module for testing kernel module.
"""

__author__ = 'wittawat'

import autograd
import autograd.numpy as np
import matplotlib.pyplot as plt
import kgof.data as data
import kgof.density as density
import kgof.util as util
import kgof.kernel as kernel
import kgof.goftest as gof
import kgof.glo as glo
import scipy.stats as stats
import numpy.testing as testing

import unittest


class TestKGauss(unittest.TestCase):
    def setUp(self):
        pass

    def test_basic(self):
        """
        Nothing special. Just test basic things.
        """
        # sample
        n = 10
        d = 3
        with util.NumpySeedContext(seed=29):
            X = np.random.randn(n, d)*3
            k = kernel.KGauss(sigma2=1)
            K = k.eval(X, X)

            self.assertEqual(K.shape, (n, n))
            self.assertTrue(np.all(K >= 0-1e-6))
            self.assertTrue(np.all(K <= 1+1e-6), 'K not bounded by 1')

    def test_pair_gradX_Y(self):
        # sample
        n = 11
        d = 3
        with util.NumpySeedContext(seed=20):
            X = np.random.randn(n, d)*4
            Y = np.random.randn(n, d)*2
            k = kernel.KGauss(sigma2=2.1)
            # n x d
            pair_grad = k.pair_gradX_Y(X, Y)
            loop_grad = np.zeros((n, d))
            for i in range(n):
                for j in range(d):
                    loop_grad[i, j] = k.gradX_Y(X[[i], :], Y[[i], :], j)

            testing.assert_almost_equal(pair_grad, loop_grad)


    def test_gradX_y(self):
        n = 10
        with util.NumpySeedContext(seed=10):
            for d in [1, 3]:
                y = np.random.randn(d)*2
                X = np.random.rand(n, d)*3

                sigma2 = 1.3
                k = kernel.KGauss(sigma2=sigma2)
                # n x d
                G = k.gradX_y(X, y)
                # check correctness 
                K = k.eval(X, y[np.newaxis, :])
                myG = -K/sigma2*(X-y)

                self.assertEqual(G.shape, myG.shape)
                testing.assert_almost_equal(G, myG)


    def test_gradXY_sum(self):
        n = 11
        with util.NumpySeedContext(seed=12):
            for d in [3, 1]:
                X = np.random.randn(n, d)
                sigma2 = 1.4
                k = kernel.KGauss(sigma2=sigma2)

                # n x n
                myG = np.zeros((n, n))
                K = k.eval(X, X)
                for i in range(n):
                    for j in range(n):
                        diffi2 = np.sum( (X[i, :] - X[j, :])**2 )
                        #myG[i, j] = -diffi2*K[i, j]/(sigma2**2)+ d*K[i, j]/sigma2
                        myG[i, j] = K[i, j]/sigma2*(d - diffi2/sigma2)

                # check correctness 
                G = k.gradXY_sum(X, X)

                self.assertEqual(G.shape, myG.shape)
                testing.assert_almost_equal(G, myG)


    def tearDown(self):
        pass


if __name__ == '__main__':
   unittest.main()




import warnings
from typing import Optional

import numpy as np
import torch
from torch.distributions import Multinomial


def get_default_device() -> torch.device:
    device = torch.ones((1,)).device
    return device


def sample_with_weights(
    values: torch.Tensor, weights: torch.Tensor, num_samples: int
) -> torch.Tensor:
    # define multinomial with weights as probs
    multi = Multinomial(probs=weights)
    # sample num samples, with replacement
    samples = multi.sample(sample_shape=(num_samples,))
    # get indices of success trials
    indices = torch.where(samples)[1]
    # return those indices from trace
    return values[indices]


def sample(
    a: torch.Tensor, num_samples: int, replace: bool = True, seed: Optional[int] = None
) -> torch.Tensor:
    """Sample with or without replacement

    NOTE: See also https://github.com/pytorch/pytorch/issues/16897
    TODO: Dispatch to `torch_sample`?
    """
    if seed is not None:
        np.random.seed(seed)
    num_elements = int(a.shape[0])
    idxs = np.random.choice(num_elements, size=(num_samples,), replace=replace)
    return a[idxs, :]


def choice(*args, **kwargs):
    """Tries to use efficient choice function if available

    Installation:
        pip install "git+https://github.com/LeviViana/torch_sampling#egg=torch_sampling"
    """
    try:
        from torch_sampling import choice

        return choice(*args, **kwargs).long()
    except:
        warnings.warn("Using numpy.random.choice.")
        return choice_numpy(*args, **kwargs).long()


def choice_numpy(
    a, num_samples: int, replace: bool = True, p: torch.Tensor = None
) -> torch.Tensor:
    """Wrap np.random.choice."""

    # Numpy choice needs probs normalized.
    if p is not None:
        if isinstance(p, torch.Tensor):
            p = p.numpy()

        p /= p.sum()

    if isinstance(a, torch.Tensor):
        a = a.numpy()

    samples = np.random.choice(a, num_samples, replace, p)

    return torch.as_tensor(samples, dtype=torch.float32)



from typing import Optional

import numpy as np
import pyabcranger
import torch
from sbi.simulators.simutils import simulate_in_batches
from tqdm import tqdm

import sbibm
from sbibm.tasks.task import Task
from sbibm.utils.torch import sample_with_weights

from .abcranger_utils import estimparam_args


def run(
    task: Task,
    num_samples: int,
    num_simulations: int,
    num_observation: Optional[int] = None,
    observation: Optional[torch.Tensor] = None,
    batch_size: int = 1_000,
    ntrees: int = 500,
    nthreads: int = 1,
) -> (torch.Tensor, int, Optional[torch.Tensor]):
    """Random forest ABC using the abcranger toolbox.

    Args:
        task: Task instance
        num_samples: Number of samples to generate from posterior
        num_simulations: Simulation budget
        num_observation: Observation number to load, alternative to `observation`
        observation: Observation, alternative to `num_observation`
        ntrees: Number of random forest trees
        nthreads: Number of cores for random forest regression

    Returns:
        Samples from posterior, number of simulator calls, log probability of true params if computable
    """
    assert not (num_observation is None and observation is None)
    assert not (num_observation is not None and observation is not None)

    log = sbibm.get_logger(__name__)
    log.info(f"Starting to run RF-ABC")

    prior = task.get_prior()
    simulator = task.get_simulator()
    if observation is None:
        observation = task.get_observation(num_observation)

    # Simulate training data set
    log.info(f"Generating data set as reference table")
    thetas = prior(num_samples=num_simulations)

    xs = simulate_in_batches(
        simulator,
        theta=thetas,
        sim_batch_size=batch_size,
        num_workers=1,
        show_progress_bars=True,
    )

    assert not thetas.isnan().any()
    assert not xs.isnan().any()
    assert not observation.isnan().any()

    dim_thetas = thetas.shape[1]
    dim_xs = xs.shape[1]

    names_thetas = [f"t{i}" for i in range(dim_thetas)]
    names_xs = [f"x{i}" for i in range(dim_xs)]

    np_thetas = thetas.numpy().astype(np.float64)
    np_xs = xs.numpy().astype(np.float64)
    np_xo = observation.reshape(-1).numpy().astype(np.float64)

    # Put data in reference table
    reftable = pyabcranger.reftable(
        num_simulations,
        # nrecscen, sth with number of scenarios, would stick with 0.
        [0],
        [dim_thetas],  # Number of parameters per model, thus the list.
        names_thetas,  # param names
        names_xs,  # data dim names
        np_xs,  # data
        np_thetas,  # params
        # called scenarios internally, maybe the indices for the models.
        np.ones(num_simulations),
    )

    # Inference per dimension
    log.info(f"Running RF inference for each parameter separately")
    postres = [
        pyabcranger.estimparam(
            reftable,  # data
            np_xo,  # observation
            estimparam_args(
                dim, num_simulations, ntrees, nthreads
            ),  # options for cpp routines as string
            True,  # verbose flag (?)
            False,  # flag for weights: whether to return the weights for oob samples (?) or not.
        )
        for dim in tqdm(range(dim_thetas))
    ]

    # Get weights per dimension
    log.info(f"Generating independent posterior samples per parameter")
    samples = []
    for dim in range(dim_thetas):
        samples_dim = np.asanyarray(postres[dim].values_weights)[:, 0]
        weights_dim = np.asanyarray(postres[dim].values_weights)[:, 1]

        samples.append(
            sample_with_weights(
                values=torch.from_numpy(samples_dim),
                weights=torch.from_numpy(weights_dim),
                num_samples=num_samples,
            )
        )

    return torch.stack(samples, dim=1), simulator.num_simulations, None



import logging

import elfi
import numpy as np

from sbibm.tasks.task import Task


def build_prior(task: Task, model: elfi.ElfiModel):
    log = logging.getLogger(__name__)
    log.warn("Will discard any correlations in prior")

    bounds = {}

    prior_cls = str(task.prior_dist)
    if prior_cls == "Independent()":
        prior_cls = str(task.prior_dist.base_dist)

    prior_params = {}
    if "MultivariateNormal" in prior_cls:
        prior_params["m"] = task.prior_params["loc"].numpy()
        if "precision_matrix" in prior_cls:
            prior_params["C"] = np.linalg.inv(
                task.prior_params["precision_matrix"].numpy()
            )
        if "covariance_matrix" in prior_cls:
            prior_params["C"] = task.prior_params["covariance_matrix"].numpy()

        for dim in range(task.dim_parameters):
            loc = prior_params["m"][dim]
            scale = np.sqrt(prior_params["C"][dim, dim])

            elfi.Prior(
                "norm", loc, scale, model=model, name=f"parameter_{dim}",
            )

            bounds[f"parameter_{dim}"] = (
                prior_params["m"][dim] - 3.0 * np.sqrt(prior_params["C"][dim, dim]),
                prior_params["m"][dim] + 3.0 * np.sqrt(prior_params["C"][dim, dim]),
            )

    elif "Uniform" in prior_cls:
        prior_params["low"] = task.prior_params["low"].numpy()
        prior_params["high"] = task.prior_params["high"].numpy()

        for dim in range(task.dim_parameters):
            loc = prior_params["low"][dim]
            scale = prior_params["high"][dim] - loc

            elfi.Prior(
                "uniform", loc, scale, model=model, name=f"parameter_{dim}",
            )

            bounds[f"parameter_{dim}"] = (
                prior_params["low"][dim],
                prior_params["high"][dim],
            )

    else:
        log.info("No support for prior yet")
        raise NotImplementedError

    return bounds



import subprocess
from pathlib import Path
from typing import Optional


def commit_sha() -> Optional[str]:
    """Get current commit SHA

    Returns:
        Commit SHA if available
    """
    try:
        commit_sha = (
            subprocess.check_output(
                ["git", "rev-parse", "--short", "HEAD"], cwd=Path(__file__).parents[1]
            )
            .decode("ascii")
            .strip()
        )
    except Exception:
        commit_sha = None

    return commit_sha



"""
Module containing implementations of some unnormalized probability density 
functions.
"""
from __future__ import division

from builtins import range
from past.utils import old_div
from builtins import object
from future.utils import with_metaclass
__author__ = 'wittawat'

from abc import ABCMeta, abstractmethod
import autograd
import autograd.numpy as np
import kgof.data as data
import scipy.stats as stats
#import warnings
import logging

def warn_bounded_domain(self):
    logging.warning('{} has a bounded domain. This may have an unintended effect to the test result of FSSD.'.format(self.__class__) )

def from_log_den(d, f):
    """
    Construct an UnnormalizedDensity from the function f, implementing the log 
    of an unnormalized density.

    f: X -> den where X: n x d and den is a numpy array of length n.
    """
    return UDFromCallable(d, flog_den=f)

def from_grad_log(d, g):
    """
    Construct an UnnormalizedDensity from the function g, implementing the
    gradient of the log of an unnormalized density.

    g: X -> grad where X: n x d and grad is n x d (2D numpy array)
    """
    return UDFromCallable(d, fgrad_log=g)


class UnnormalizedDensity(with_metaclass(ABCMeta, object)):
    """
    An abstract class of an unnormalized probability density function.  This is
    intended to be used to represent a model of the data for goodness-of-fit
    testing.
    """

    @abstractmethod
    def log_den(self, X):
        """
        Evaluate this log of the unnormalized density on the n points in X.

        X: n x d numpy array

        Return a one-dimensional numpy array of length n.
        """
        raise NotImplementedError()

    def log_normalized_den(self, X):
        """
        Evaluate the exact normalized log density. The difference to log_den()
        is that this method adds the normalizer. This method is not
        compulsory. Subclasses do not need to override.
        """
        raise NotImplementedError()

    def get_datasource(self):
        """
        Return a DataSource that allows sampling from this density.
        May return None if no DataSource is implemented.
        Implementation of this method is not enforced in the subclasses.
        """
        return None

    def grad_log(self, X):
        """
        Evaluate the gradients (with respect to the input) of the log density at
        each of the n points in X. This is the score function. Given an
        implementation of log_den(), this method will automatically work.
        Subclasses may override this if a more efficient implementation is
        available.

        X: n x d numpy array.

        Return an n x d numpy array of gradients.
        """
        g = autograd.elementwise_grad(self.log_den)
        G = g(X)
        return G

    @abstractmethod
    def dim(self):
        """
        Return the dimension of the input.
        """
        raise NotImplementedError()

# end UnnormalizedDensity

class UDFromCallable(UnnormalizedDensity):
    """
    UnnormalizedDensity constructed from the specified implementations of 
    log_den() and grad_log() as callable objects.
    """
    def __init__(self, d, flog_den=None, fgrad_log=None):
        """
        Only one of log_den and grad_log are required.
        If log_den is specified, the gradient is automatically computed with
        autograd.

        d: the dimension of the domain of the density
        log_den: a callable object (function) implementing the log of an unnormalized density. See UnnormalizedDensity.log_den.
        grad_log: a callable object (function) implementing the gradient of the log of an unnormalized density.
        """
        if flog_den is None and fgrad_log is None:
            raise ValueError('At least one of {log_den, grad_log} must be specified.')
        self.d = d
        self.flog_den = flog_den
        self.fgrad_log = fgrad_log

    def log_den(self, X):
        flog_den = self.flog_den
        if flog_den is None:
            raise ValueError('log_den callable object is None.')
        return flog_den(X)

    def grad_log(self, X):
        fgrad_log = self.fgrad_log
        if fgrad_log is None:
            # autograd
            g = autograd.elementwise_grad(self.flog_den)
            G = g(X)
        else:
            G = fgrad_log(X)
        return G

    def dim(self):
        return self.d

# end UDFromCallable


class IsotropicNormal(UnnormalizedDensity):
    """
    Unnormalized density of an isotropic multivariate normal distribution.
    """
    def __init__(self, mean, variance):
        """
        mean: a numpy array of length d for the mean 
        variance: a positive floating-point number for the variance.
        """
        self.mean = mean 
        self.variance = variance

    def log_den(self, X):
        mean = self.mean 
        variance = self.variance
        unden = old_div(-np.sum((X-mean)**2, 1),(2.0*variance))
        return unden

    def log_normalized_den(self, X):
        d = self.dim()
        return stats.multivariate_normal.logpdf(X, mean=self.mean, cov=self.variance*np.eye(d))

    def get_datasource(self):
        return data.DSIsotropicNormal(self.mean, self.variance)

    def dim(self):
        return len(self.mean)



class Normal(UnnormalizedDensity):
    """
    A multivariate normal distribution.
    """
    def __init__(self, mean, cov):
        """
        mean: a numpy array of length d.
        cov: d x d numpy array for the covariance.
        """
        self.mean = mean 
        self.cov = cov
        assert mean.shape[0] == cov.shape[0]
        assert cov.shape[0] == cov.shape[1]
        E, V = np.linalg.eigh(cov)
        if np.any(np.abs(E) <= 1e-7):
            raise ValueError('covariance matrix is not full rank.')
        # The precision matrix
        self.prec = np.dot(np.dot(V, np.diag(old_div(1.0,E))), V.T)
        #print self.prec

    def log_den(self, X):
        mean = self.mean 
        X0 = X - mean
        X0prec = np.dot(X0, self.prec)
        unden = old_div(-np.sum(X0prec*X0, 1),2.0)
        return unden

    def get_datasource(self):
        return data.DSNormal(self.mean, self.cov)

    def dim(self):
        return len(self.mean)

# end Normal

class IsoGaussianMixture(UnnormalizedDensity):
    """
    UnnormalizedDensity of a Gaussian mixture in R^d where each component 
    is an isotropic multivariate normal distribution.

    Let k be the number of mixture components.
    """
    def __init__(self, means, variances, pmix=None):
        """
        means: a k x d 2d array specifying the means.
        variances: a one-dimensional length-k array of variances
        pmix: a one-dimensional length-k array of mixture weights. Sum to one.
        """
        k, d = means.shape
        if k != len(variances):
            raise ValueError('Number of components in means and variances do not match.')

        if pmix is None:
            pmix = old_div(np.ones(k),float(k))

        if np.abs(np.sum(pmix) - 1) > 1e-8:
            raise ValueError('Mixture weights do not sum to 1.')

        self.pmix = pmix
        self.means = means
        self.variances = variances

    def log_den(self, X):
        return self.log_normalized_den(X)

    def log_normalized_den(self, X):
        pmix = self.pmix
        means = self.means
        variances = self.variances
        k, d = self.means.shape
        n = X.shape[0]
        den = np.zeros(n, dtype=float)
        for i in range(k):
            norm_den_i = IsoGaussianMixture.normal_density(means[i],
                    variances[i], X)
            den = den + norm_den_i*pmix[i]
        return np.log(den)

 
    #def grad_log(self, X):
    #    """
    #    Return an n x d numpy array of gradients.
    #    """
    #    pmix = self.pmix
    #    means = self.means
    #    variances = self.variances
    #    k, d = self.means.shape
    #    # exact density. length-n array
    #    den = np.exp(self.log_den(X))
    #    for i in range(k):
    #        norm_den_i = IsoGaussianMixture.normal_density(means[i],
    #                variances[i], X)


    @staticmethod
    def normal_density(mean, variance, X):
        """
        Exact density (not log density) of an isotropic Gaussian.
        mean: length-d array
        variance: scalar variances
        X: n x d 2d-array
        """
        Z = np.sqrt(2.0*np.pi*variance)
        unden = np.exp(old_div(-np.sum((X-mean)**2.0, 1),(2.0*variance)) )
        den = old_div(unden,Z)
        assert len(den) == X.shape[0]
        return den

    def get_datasource(self):
        return data.DSIsoGaussianMixture(self.means, self.variances, self.pmix)

    def dim(self):
        k, d = self.means.shape
        return d

# end class IsoGaussianMixture

class GaussianMixture(UnnormalizedDensity):
    """
    UnnormalizedDensity of a Gaussian mixture in R^d where each component 
    can be arbitrary. This is the most general form of a Gaussian mixture.

    Let k be the number of mixture components.
    """
    def __init__(self, means, variances, pmix=None):
        """
        means: a k x d 2d array specifying the means.
        variances: a k x d x d numpy array containing a stack of k covariance
            matrices, one for each mixture component.
        pmix: a one-dimensional length-k array of mixture weights. Sum to one.
        """
        k, d = means.shape
        if k != variances.shape[0]:
            raise ValueError('Number of components in means and variances do not match.')

        if pmix is None:
            pmix = old_div(np.ones(k),float(k))

        if np.abs(np.sum(pmix) - 1) > 1e-8:
            raise ValueError('Mixture weights do not sum to 1.')

        self.pmix = pmix
        self.means = means
        self.variances = variances

    def log_den(self, X):
        return self.log_normalized_den(X)

    def log_normalized_den(self, X):
        pmix = self.pmix
        means = self.means
        variances = self.variances
        k, d = self.means.shape
        n = X.shape[0]

        den = np.zeros(n, dtype=float)
        for i in range(k):
            norm_den_i = GaussianMixture.multivariate_normal_density(means[i],
                    variances[i], X)
            den = den + norm_den_i*pmix[i]
        return np.log(den)

    @staticmethod
    def multivariate_normal_density(mean, cov, X):
        """
        Exact density (not log density) of a multivariate Gaussian.
        mean: length-d array
        cov: a dxd covariance matrix
        X: n x d 2d-array
        """
        
        evals, evecs = np.linalg.eigh(cov)
        cov_half_inv = evecs.dot(np.diag(evals**(-0.5))).dot(evecs.T)
    #     print(evals)
        half_evals = np.dot(X-mean, cov_half_inv)
        full_evals = np.sum(half_evals**2, 1)
        unden = np.exp(-0.5*full_evals)
        
        Z = np.sqrt(np.linalg.det(2.0*np.pi*cov))
        den = unden/Z
        assert len(den) == X.shape[0]
        return den

    def get_datasource(self):
        return data.DSGaussianMixture(self.means, self.variances, self.pmix)

    def dim(self):
        k, d = self.means.shape
        return d

# end GaussianMixture

class GaussBernRBM(UnnormalizedDensity):
    """
    Gaussian-Bernoulli Restricted Boltzmann Machine.
    The joint density takes the form
        p(x, h) = Z^{-1} exp(0.5*x^T B h + b^T x + c^T h - 0.5||x||^2)
    where h is a vector of {-1, 1}.
    """
    def __init__(self, B, b, c):
        """
        B: a dx x dh matrix 
        b: a numpy array of length dx
        c: a numpy array of length dh
        """
        dh = len(c)
        dx = len(b)
        assert B.shape[0] == dx
        assert B.shape[1] == dh
        assert dx > 0
        assert dh > 0
        self.B = B
        self.b = b
        self.c = c

    def log_den(self, X):
        B = self.B
        b = self.b
        c = self.c

        XBC = 0.5*np.dot(X, B) + c
        unden = np.dot(X, b) - 0.5*np.sum(X**2, 1) + np.sum(np.log(np.exp(XBC)
            + np.exp(-XBC)), 1)
        assert len(unden) == X.shape[0]
        return unden

    def grad_log(self, X):
    #    """
    #    Evaluate the gradients (with respect to the input) of the log density at
    #    each of the n points in X. This is the score function.

    #    X: n x d numpy array.
        """
        Evaluate the gradients (with respect to the input) of the log density at
        each of the n points in X. This is the score function.

        X: n x d numpy array.

        Return an n x d numpy array of gradients.
        """
        XB = np.dot(X, self.B)
        Y = 0.5*XB + self.c
        E2y = np.exp(2*Y)
        # n x dh
        Phi = old_div((E2y-1.0),(E2y+1))
        # n x dx
        T = np.dot(Phi, 0.5*self.B.T)
        S = self.b - X + T
        return S

    def get_datasource(self, burnin=2000):
        return data.DSGaussBernRBM(self.B, self.b, self.c, burnin=burnin)

    def dim(self):
        return len(self.b)

# end GaussBernRBM

class ISIPoissonLinear(UnnormalizedDensity):
    """
    Unnormalized density of inter-arrival times from nonhomogeneous poisson process with linear intensity function.
    lambda = 1 + bt
    """
    def __init__(self, b):
        """
        b: slope of the linear function 
        """
        warn_bounded_domain(self)
        self.b = b 
    
    def log_den(self, X):
        b = self.b
        unden = -np.sum(0.5*b*X**2+X-np.log(1.0+b*X), 1)
        return unden

    def dim(self):
        return 1

# end ISIPoissonLinear

class ISIPoissonSine(UnnormalizedDensity):
    """
    Unnormalized density of inter-arrival times from nonhomogeneous poisson process with sine intensity function.
    lambda = b*(1+sin(w*X))
    """
    def __init__(self, w=10.0,b=1.0):
        """
        w: the frequency of sine function
        b: amplitude of intensity function
        """
        warn_bounded_domain(self)
        self.b = b
        self.w = w
    
    def log_den(self, X):
        b = self.b
        w = self.w
        unden = np.sum(b*(-X + old_div((np.cos(w*X)-1),w)) + np.log(b*(1+np.sin(w*X))),1)
        return unden

    def dim(self):
        return 1

# end ISIPoissonSine

class Gamma(UnnormalizedDensity):
    """
    A gamma distribution.
    """
    def __init__(self, alpha, beta = 1.0):
        """
        alpha: shape of parameter
        beta: scale
        """
        warn_bounded_domain(self)
        self.alpha = alpha 
        self.beta = beta
        
    def log_den(self, X):
        alpha = self.alpha
        beta = self.beta
        #unden = np.sum(stats.gamma.logpdf(X, alpha, scale = beta), 1)
        unden = np.sum(-beta*X + (alpha-1)*np.log(X), 1)
        return unden

    def get_datasource(self):
        return data.DSNormal(self.mean, self.cov)


    def dim(self):
        return 1


class LogGamma(UnnormalizedDensity):
    """
    A gamma distribution with transformed domain.
    t = exp(x),  t \in R+  x \in R
    """
    def __init__(self, alpha, beta = 1.0):
        """
        alpha: shape of parameter
        beta: scale
        """
        self.alpha = alpha
        self.beta = beta
        
    def log_den(self, X):
        alpha = self.alpha
        beta = self.beta
        #unden = np.sum(stats.gamma.logpdf(X, alpha, scale = beta), 1)
        unden = np.sum(-beta*np.exp(X) + (alpha-1)*X + X , 1)
        return unden

    def get_datasource(self):
        return data.DSNormal(self.mean, self.cov)

    def dim(self):
        return 1
# end LogGamma



class ISILogPoissonLinear(UnnormalizedDensity):
    """
    Unnormalized density of inter-arrival times from nonhomogeneous poisson process with linear intensity function.
    lambda = 1 + bt
    """
    def __init__(self, b):
        """
        b: slope of the linear function 
        """
        warn_bounded_domain(self)
        self.b = b 
    
    def log_den(self, X):
        b = self.b
        unden = -np.sum(0.5*b*np.exp(X)**2 + np.exp(X) - np.log(1.0+b*np.exp(X))-X, 1)
        return unden

    def dim(self):
        return 1

# end ISIPoissonLinear

class ISIPoisson2D(UnnormalizedDensity):
    """
    Unnormalized density of nonhomogeneous spatial poisson process
    """
    def __init__(self):
        """
        lambda_(X,Y) = X^2 + Y^2
        """
        warn_bounded_domain(self)

    def quadratic_intensity(self,X,Y):
        int_intensity = -(X**2+Y**2)*X*Y + 3*np.log(X**2+Y**2)
        return int_intensity

    def log_den(self, X):
        unden = self.quadratic_intensity(X[:,0],X[:,1])
        return unden

    def dim(self):
        return 1

# end class ISIPoisson2D


class ISISigmoidPoisson2D(UnnormalizedDensity):
    """
    Unnormalized density of nonhomogeneous spatial poisson process with sigmoid transformation
    """
    def __init__(self, intensity = 'quadratic', w = 1.0, a=1.0):
        """
        lambda_(X,Y) = a* X^2 + Y^2
        X = 1/(1+exp(s))
        Y = 1/(1+exp(t))
        X, Y \in [0,1], s,t \in R
        """
        warn_bounded_domain(self)
        self.a = a
        self.w = w
        if intensity == 'quadratic':
            self.intensity = self.quadratic_intensity
        elif intensity == 'sine':
            self.intensity = self.sine_intensity
        else:
            raise ValueError('Not intensity function found')

    def sigmoid(self,x):
        sig = old_div(1,(1+np.exp(x)))
        return sig

    def quadratic_intensity(self,s,t):
        X = self.sigmoid(s)
        Y = self.sigmoid(t)
        int_intensity = -(self.a*X**2+Y**2)*X*Y + 3*(np.log(self.a*X**2+Y**2)+np.log((X*(X-1)*Y*(Y-1))))
        return int_intensity

    def log_den(self, S):
        unden = self.quadratic_intensity(S[:,0],S[:,1])
        return unden

    def dim(self):
        return 1

# end class ISISigmoidPoisson2D


class Poisson2D(UnnormalizedDensity):
    """
    Unnormalized density of nonhomogeneous spatial poisson process
    """
    def __init__(self, w=1.0):
        """
        lambda_(X,Y) = sin(w*pi*X)+sin(w*pi*Y)
        """
        self.w = w

    def lamb_sin(self, X):
        return np.prod(np.sin(self.w*np.pi*X),1)

    def log_den(self, X):
        unden = np.log(self.gmm_den(X))
        return unden

    def dim(self):
        return 1
    
class Resample(UnnormalizedDensity):
    """
    Unnormalized Density of real dataset with estimated intensity function
    fit takes the function to evaluate the density of resampled data
    """
    def __init__(self, fit):
        self.fit = fit

    def log_den(self, X):
        unden = np.log(self.fit(X))
        return unden

    def dim(self):
        return 1

# end class SigmoidPoisson2D

class GaussCosFreqs(UnnormalizedDensity):
    """
    p(x) \propto exp(-||x||^2/2sigma^2)*(1+ prod_{i=1}^d cos(w_i*x_i))

    where w1,..wd are frequencies of each dimension.
    sigma^2 is the overall variance.
    """

    def __init__(self, sigma2, freqs):
        """
        sigma2: overall scale of the distribution. A positive scalar.
        freqs: a 1-d array of length d for the frequencies.
        """
        self.sigma2 = sigma2
        if sigma2 <= 0 :
            raise ValueError('sigma2 must be > 0')
        self.freqs = freqs

    def log_den(self, X):
        sigma2 = self.sigma2
        freqs = self.freqs
        log_unden = old_div(-np.sum(X**2, 1),(2.0*sigma2)) + 1+np.prod(np.cos(X*freqs), 1)
        return log_unden

    def dim(self):
        return len(self.freqs)

    def get_datasource(self):
        return data.DSGaussCosFreqs(self.sigma2, self.freqs)




import logging
import shutil
from pathlib import Path
from typing import Any, Callable

import torch
from sbi.utils.plot import pairplot
from torch.utils.tensorboard import SummaryWriter


def tb_plot_posterior(
    writer: SummaryWriter, samples: torch.Tensor, tag: str = "posterior"
):
    if type(samples) == torch.Tensor:
        samples = samples.numpy()
    fig, _ = pairplot(samples.squeeze(), points=[])
    writer.add_figure(f"{tag}", fig, close=True)


def tb_make_writer(
    logger: logging.Logger = None, basepath: str = "tensorboard",
) -> (SummaryWriter, Callable):
    """Builds tensorboard summary writers
    """
    log_dir = Path(f"{basepath}/summary")
    if log_dir.exists() and log_dir.is_dir():
        shutil.rmtree(log_dir)
    writer = SummaryWriter(log_dir=log_dir)

    if logger is not None:
        tbh = TensorboardHandler(writer)
        logger.addHandler(tbh)

    def close_fn():
        writer.flush()
        writer.close()
        logger.removeHandler(tbh)

    return writer, close_fn


class TensorboardHandler(logging.Handler):
    def __init__(
        self, writer: SummaryWriter, *args: Any, **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.writer = writer

    def emit(self, record: logging.LogRecord):
        try:
            self.writer.add_text("logging", str(record.msg))
            self.writer.flush()
        except:
            print(f"Could not log {record} with msg: {record.msg}")



from sbibm.algorithms.sbi.mcabc import run as mcabc
from sbibm.algorithms.sbi.smcabc import run as smcabc
from sbibm.algorithms.sbi.snle import run as snle
from sbibm.algorithms.sbi.snpe import run as snpe
from sbibm.algorithms.sbi.snre import run as snre



from typing import Any

import torch
from tqdm.auto import tqdm

from sbibm.tasks.task import Task


def run(
    task: Task,
    num_samples: int,
    num_simulations: int,
    batch_size: int = 1,
    **kwargs: Any,
) -> torch.Tensor:
    """Runtime baseline

    Draws `num_simulations` samples from prior and simulates, discards outcomes,
    returns tensor of NaNs.

    Args:
        task: Task instance
        num_samples: Number of samples to generate from posterior
        num_simulations: Simulation budget
        batch_size: Batch size for simulations

    Returns:
        Random samples from prior
    """
    prior = task.get_prior()
    simulator = task.get_simulator()

    batch_size = min(batch_size, num_simulations)
    num_batches = int(num_simulations / batch_size)

    for i in tqdm(range(num_batches)):
        _ = simulator(prior(num_samples=batch_size))

    assert simulator.num_simulations == num_simulations

    samples = float("nan") * torch.ones((num_samples, task.dim_parameters))

    return samples



"""Simulation to examine the P(reject) as the parameters for each problem are 
varied. What varies will depend on the problem."""

__author__ = 'wittawat'

import kgof
import kgof.data as data
import kgof.glo as glo
import kgof.density as density
import kgof.goftest as gof
import kgof.intertst as tgof
import kgof.mmd as mgof
import kgof.util as util 
import kgof.kernel as kernel 

# need independent_jobs package 
# https://github.com/karlnapf/independent-jobs
# The independent_jobs and kgof have to be in the global search path (.bashrc)
import independent_jobs as inj
from independent_jobs.jobs.IndependentJob import IndependentJob
from independent_jobs.results.SingleResult import SingleResult
from independent_jobs.aggregators.SingleResultAggregator import SingleResultAggregator
from independent_jobs.engines.BatchClusterParameters import BatchClusterParameters
from independent_jobs.engines.SerialComputationEngine import SerialComputationEngine
from independent_jobs.engines.SlurmComputationEngine import SlurmComputationEngine
from independent_jobs.tools.Log import logger
import logging
import math
#import numpy as np
import autograd.numpy as np
import os
import sys 
import time

"""
All the job functions return a dictionary with the following keys:
    - goftest: test object. (may or may not return)
    - test_result: the result from calling perform_test(te).
    - time_secs: run time in seconds 
"""

def job_fssdJ1q_med(p, data_source, tr, te, r, J=1, null_sim=None):
    """
    FSSD test with a Gaussian kernel, where the test locations are randomized,
    and the Gaussian width is set with the median heuristic. Use full sample.
    No training/testing splits.

    p: an UnnormalizedDensity
    data_source: a DataSource
    tr, te: Data
    r: trial number (positive integer)
    """
    if null_sim is None:
        null_sim = gof.FSSDH0SimCovObs(n_simulate=2000, seed=r)

    # full data
    data = tr + te
    X = data.data()
    with util.ContextTimer() as t:
        # median heuristic 
        med = util.meddistance(X, subsample=1000)
        k = kernel.KGauss(med**2)
        V = util.fit_gaussian_draw(X, J, seed=r+3)

        fssd_med = gof.FSSD(p, k, V, null_sim=null_sim, alpha=alpha)
        fssd_med_result = fssd_med.perform_test(data)
    return {
            'goftest': fssd_med,
            'test_result': fssd_med_result, 'time_secs': t.secs}


def job_fssdJ5q_med(p, data_source, tr, te, r):
    """
    FSSD. J=5
    """
    return job_fssdJ1q_med(p, data_source, tr, te, r, J=5)


def job_fssdJ1q_opt(p, data_source, tr, te, r, J=1, null_sim=None):
    """
    FSSD with optimization on tr. Test on te. Use a Gaussian kernel.
    """
    if null_sim is None:
        null_sim = gof.FSSDH0SimCovObs(n_simulate=2000, seed=r)

    Xtr = tr.data()
    with util.ContextTimer() as t:
        # Use grid search to initialize the gwidth
        n_gwidth_cand = 5
        gwidth_factors = 2.0**np.linspace(-3, 3, n_gwidth_cand) 
        med2 = util.meddistance(Xtr, 1000)**2

        k = kernel.KGauss(med2*2)
        # fit a Gaussian to the data and draw to initialize V0
        V0 = util.fit_gaussian_draw(Xtr, J, seed=r+1, reg=1e-6)
        list_gwidth = np.hstack( ( (med2)*gwidth_factors ) )
        besti, objs = gof.GaussFSSD.grid_search_gwidth(p, tr, V0, list_gwidth)
        gwidth = list_gwidth[besti]
        assert util.is_real_num(gwidth), 'gwidth not real. Was %s'%str(gwidth)
        assert gwidth > 0, 'gwidth not positive. Was %.3g'%gwidth
        logging.info('After grid search, gwidth=%.3g'%gwidth)
        
        ops = {
            'reg': 1e-2,
            'max_iter': 40,
            'tol_fun': 1e-4,
            'disp': True,
            'locs_bounds_frac':10.0,
            'gwidth_lb': 1e-1,
            'gwidth_ub': 1e4,
            }

        V_opt, gwidth_opt, info = gof.GaussFSSD.optimize_locs_widths(p, tr,
                gwidth, V0, **ops) 
        # Use the optimized parameters to construct a test
        k_opt = kernel.KGauss(gwidth_opt)
        fssd_opt = gof.FSSD(p, k_opt, V_opt, null_sim=null_sim, alpha=alpha)
        fssd_opt_result = fssd_opt.perform_test(te)
    return {'test_result': fssd_opt_result, 'time_secs': t.secs, 
            'goftest': fssd_opt, 'opt_info': info,
            }

def job_fssdJ5q_opt(p, data_source, tr, te, r):
    return job_fssdJ1q_opt(p, data_source, tr, te, r, J=5)

def job_fssdJ10q_opt(p, data_source, tr, te, r):
    return job_fssdJ1q_opt(p, data_source, tr, te, r, J=10)

def job_fssdJ5p_opt(p, data_source, tr, te, r):
    """
    The suffix p means that p is sampled to get a sample for computing the
    covariance matrix under H0.
    """
    null_sim = gof.FSSDH0SimCovDraw(n_draw=2000, n_simulate=2000, seed=r)
    return job_fssdJ1q_opt(p, data_source, tr, te, r, J=5, null_sim=null_sim)

def job_fssdJ10p_opt(p, data_source, tr, te, r):
    """
    The suffix p means that p is sampled to get a sample for computing the
    covariance matrix under H0.
    """
    null_sim = gof.FSSDH0SimCovDraw(n_draw=2000, n_simulate=2000, seed=r)
    return job_fssdJ1q_opt(p, data_source, tr, te, r, J=10, null_sim=null_sim)

def job_fssdJ1q_imq_optv(p, data_source, tr, te, r, J=1, b=-0.5, null_sim=None):
    """
    FSSD with optimization on tr. Test on te. Use an inverse multiquadric
    kernel (IMQ). Optimize only the test locations (V). Fix the kernel
    parameters to b = -0.5, c=1. These are the recommended values from

        Measuring Sample Quality with Kernels
        Jackson Gorham, Lester Mackey
    """
    if null_sim is None:
        null_sim = gof.FSSDH0SimCovObs(n_simulate=2000, seed=r)

    Xtr = tr.data()
    with util.ContextTimer() as t:
        # IMQ kernel parameters: b and c
        c = 1.0

        # fit a Gaussian to the data and draw to initialize V0
        V0 = util.fit_gaussian_draw(Xtr, J, seed=r+1, reg=1e-6)

        ops = {
            'reg': 1e-5,
            'max_iter': 30,
            'tol_fun': 1e-6,
            'disp': True,
            'locs_bounds_frac':20.0,
            }

        V_opt, info = gof.IMQFSSD.optimize_locs(p, tr, b, c, V0, **ops) 

        k_imq = kernel.KIMQ(b=b, c=c)

        # Use the optimized parameters to construct a test
        fssd_imq = gof.FSSD(p, k_imq, V_opt, null_sim=null_sim, alpha=alpha)
        fssd_imq_result = fssd_imq.perform_test(te)

    return {'test_result': fssd_imq_result, 'time_secs': t.secs, 
            'goftest': fssd_imq, 'opt_info': info,
            }

def job_fssdJ5q_imq_optv(p, data_source, tr, te, r):
    return job_fssdJ1q_imq_optv(p, data_source, tr, te, r, J=5)

def job_fssdJ5q_imqb1_optv(p, data_source, tr, te, r):
    return job_fssdJ1q_imq_optv(p, data_source, tr, te, r, J=5, b=-1.0)

def job_fssdJ5q_imqb2_optv(p, data_source, tr, te, r):
    return job_fssdJ1q_imq_optv(p, data_source, tr, te, r, J=5, b=-2.0)

def job_fssdJ5q_imqb3_optv(p, data_source, tr, te, r):
    return job_fssdJ1q_imq_optv(p, data_source, tr, te, r, J=5, b=-3.0)

def job_fssdJ1q_imq_opt(p, data_source, tr, te, r, J=1, null_sim=None):
    """
    FSSD with optimization on tr. Test on te. Use an inverse multiquadric
    kernel (IMQ). Optimize all parameters: the test locations (V), b and c (in
    the kernel).
    """
    if null_sim is None:
        null_sim = gof.FSSDH0SimCovObs(n_simulate=2000, seed=r)

    Xtr = tr.data()
    with util.ContextTimer() as t:
        # Initial IMQ kernel parameters: b and c
        b0 = -0.5
        c0 = 1.0

        # fit a Gaussian to the data and draw to initialize V0
        V0 = util.fit_gaussian_draw(Xtr, J, seed=r+1, reg=1e-6)

        ops = {
            'reg': 1e-5,
            'max_iter': 50,
            'tol_fun': 1e-6,
            'disp': True,
            'locs_bounds_frac':20.0,
            # IMQ kernel bounds
            'b_lb': -3,
            'b_ub': -0.5,
            'c_lb': 1e-1,
            'c_ub': np.sqrt(10),
            }

        V_opt, b_opt, c_opt, info = gof.IMQFSSD.optimize_locs_params(p, tr, b0,
                c0, V0, **ops) 

        k_imq = kernel.KIMQ(b=b_opt, c=c_opt)

        # Use the optimized parameters to construct a test
        fssd_imq = gof.FSSD(p, k_imq, V_opt, null_sim=null_sim, alpha=alpha)
        fssd_imq_result = fssd_imq.perform_test(te)

    return {'test_result': fssd_imq_result, 'time_secs': t.secs, 
            'goftest': fssd_imq, 'opt_info': info,
            }

def job_fssdJ5q_imq_opt(p, data_source, tr, te, r, null_sim=None):
    return job_fssdJ1q_imq_opt(p, data_source, tr, te, r, J=5)

def job_fssdJ1q_imq_optbv(p, data_source, tr, te, r, J=1, null_sim=None):
    """
    FSSD with optimization on tr. Test on te. Use an inverse multiquadric
    kernel (IMQ). Optimize the test locations (V), and b. Fix c (in the kernel)
    """
    if null_sim is None:
        null_sim = gof.FSSDH0SimCovObs(n_simulate=2000, seed=r)

    Xtr = tr.data()
    with util.ContextTimer() as t:
        # Initial IMQ kernel parameters: b and c
        b0 = -0.5
        # Fix c to this value
        c = 1.0
        c0 = c

        # fit a Gaussian to the data and draw to initialize V0
        V0 = util.fit_gaussian_draw(Xtr, J, seed=r+1, reg=1e-6)

        ops = {
            'reg': 1e-5,
            'max_iter': 40,
            'tol_fun': 1e-6,
            'disp': True,
            'locs_bounds_frac':20.0,
            # IMQ kernel bounds
            'b_lb': -20,
            'c_lb': c,
            'c_ub': c,
            }

        V_opt, b_opt, c_opt, info = gof.IMQFSSD.optimize_locs_params(p, tr, b0,
                c0, V0, **ops) 

        k_imq = kernel.KIMQ(b=b_opt, c=c_opt)

        # Use the optimized parameters to construct a test
        fssd_imq = gof.FSSD(p, k_imq, V_opt, null_sim=null_sim, alpha=alpha)
        fssd_imq_result = fssd_imq.perform_test(te)

    return {'test_result': fssd_imq_result, 'time_secs': t.secs, 
            'goftest': fssd_imq, 'opt_info': info,
            }

def job_fssdJ5q_imq_optbv(p, data_source, tr, te, r, null_sim=None):
    return job_fssdJ1q_imq_optbv(p, data_source, tr, te, r, J=5)


def job_me_opt(p, data_source, tr, te, r, J=5):
    """
    ME test of Jitkrittum et al., 2016 used as a goodness-of-fit test.
    Gaussian kernel. Optimize test locations and Gaussian width.
    """
    data = tr + te
    X = data.data()
    with util.ContextTimer() as t:
        # median heuristic 
        #pds = p.get_datasource()
        #datY = pds.sample(data.sample_size(), seed=r+294)
        #Y = datY.data()
        #XY = np.vstack((X, Y))
        #med = util.meddistance(XY, subsample=1000)
        op = {'n_test_locs': J, 'seed': r+5, 'max_iter': 40, 
             'batch_proportion': 1.0, 'locs_step_size': 1.0, 
              'gwidth_step_size': 0.1, 'tol_fun': 1e-4, 
              'reg': 1e-4}
        # optimize on the training set
        me_opt = tgof.GaussMETestOpt(p, n_locs=J, tr_proportion=tr_proportion,
                alpha=alpha, seed=r+111)

        me_result = me_opt.perform_test(data, op)
    return { 'test_result': me_result, 'time_secs': t.secs}

def job_kstein_med(p, data_source, tr, te, r):
    """
    Kernel Stein discrepancy test of Liu et al., 2016 and Chwialkowski et al.,
    2016. Use full sample. Use Gaussian kernel.
    """
    # full data
    data = tr + te
    X = data.data()
    with util.ContextTimer() as t:
        # median heuristic 
        med = util.meddistance(X, subsample=1000)
        k = kernel.KGauss(med**2)

        kstein = gof.KernelSteinTest(p, k, alpha=alpha, n_simulate=1000, seed=r)
        kstein_result = kstein.perform_test(data)
    return { 'test_result': kstein_result, 'time_secs': t.secs}

def job_kstein_imq(p, data_source, tr, te, r):
    """
    Kernel Stein discrepancy test of Liu et al., 2016 and Chwialkowski et al.,
    2016. Use full sample. Use the inverse multiquadric kernel (IMQ) studied 
    in 

    Measuring Sample Quality with Kernels
    Gorham and Mackey 2017. 

    Parameters are fixed to the recommented values: beta = b = -0.5, c = 1. 
    """
    # full data
    data = tr + te
    X = data.data()
    with util.ContextTimer() as t:
        k = kernel.KIMQ(b=-0.5, c=1.0)

        kstein = gof.KernelSteinTest(p, k, alpha=alpha, n_simulate=1000, seed=r)
        kstein_result = kstein.perform_test(data)
    return { 'test_result': kstein_result, 'time_secs': t.secs}

def job_lin_kstein_med(p, data_source, tr, te, r):
    """
    Linear-time version of the kernel Stein discrepancy test of Liu et al.,
    2016 and Chwialkowski et al., 2016. Use full sample.
    """
    # full data
    data = tr + te
    X = data.data()
    with util.ContextTimer() as t:
        # median heuristic 
        med = util.meddistance(X, subsample=1000)
        k = kernel.KGauss(med**2)

        lin_kstein = gof.LinearKernelSteinTest(p, k, alpha=alpha, seed=r)
        lin_kstein_result = lin_kstein.perform_test(data)
    return { 'test_result': lin_kstein_result, 'time_secs': t.secs}

def job_mmd_med(p, data_source, tr, te, r):
    """
    MMD test of Gretton et al., 2012 used as a goodness-of-fit test.
    Require the ability to sample from p i.e., the UnnormalizedDensity p has 
    to be able to return a non-None from get_datasource()
    """
    # full data
    data = tr + te
    X = data.data()
    with util.ContextTimer() as t:
        # median heuristic 
        pds = p.get_datasource()
        datY = pds.sample(data.sample_size(), seed=r+294)
        Y = datY.data()
        XY = np.vstack((X, Y))

        # If p, q differ very little, the median may be very small, rejecting H0
        # when it should not?
        medx = util.meddistance(X, subsample=1000)
        medy = util.meddistance(Y, subsample=1000)
        medxy = util.meddistance(XY, subsample=1000)
        med_avg = (medx+medy+medxy)/3.0
        k = kernel.KGauss(med_avg**2)

        mmd_test = mgof.QuadMMDGof(p, k, n_permute=400, alpha=alpha, seed=r)
        mmd_result = mmd_test.perform_test(data)
    return { 'test_result': mmd_result, 'time_secs': t.secs}

def job_mmd_opt(p, data_source, tr, te, r):
    """
    MMD test of Gretton et al., 2012 used as a goodness-of-fit test.
    Require the ability to sample from p i.e., the UnnormalizedDensity p has 
    to be able to return a non-None from get_datasource()

    With optimization. Gaussian kernel.
    """
    data = tr + te
    X = data.data()
    with util.ContextTimer() as t:
        # median heuristic 
        pds = p.get_datasource()
        datY = pds.sample(data.sample_size(), seed=r+294)
        Y = datY.data()
        XY = np.vstack((X, Y))

        med = util.meddistance(XY, subsample=1000)

        # Construct a list of kernels to try based on multiples of the median
        # heuristic
        #list_gwidth = np.hstack( (np.linspace(20, 40, 10), (med**2)
        #    *(2.0**np.linspace(-2, 2, 20) ) ) )
        list_gwidth = (med**2)*(2.0**np.linspace(-3, 3, 30) ) 
        list_gwidth.sort()
        candidate_kernels = [kernel.KGauss(gw2) for gw2 in list_gwidth]

        mmd_opt = mgof.QuadMMDGofOpt(p, n_permute=300, alpha=alpha, seed=r+56)
        mmd_result = mmd_opt.perform_test(data,
                candidate_kernels=candidate_kernels,
                tr_proportion=tr_proportion, reg=1e-3)
    return { 'test_result': mmd_result, 'time_secs': t.secs}


# Define our custom Job, which inherits from base class IndependentJob
class Ex2Job(IndependentJob):
   
    def __init__(self, aggregator, p, data_source,
            prob_label, rep, job_func, prob_param):
        #walltime = 60*59*24 
        walltime = 60*59
        memory = int(tr_proportion*sample_size*1e-2) + 50

        IndependentJob.__init__(self, aggregator, walltime=walltime,
                               memory=memory)
        # p: an UnnormalizedDensity
        self.p = p
        self.data_source = data_source
        self.prob_label = prob_label
        self.rep = rep
        self.job_func = job_func
        self.prob_param = prob_param

    # we need to define the abstract compute method. It has to return an instance
    # of JobResult base class
    def compute(self):

        p = self.p
        data_source = self.data_source 
        r = self.rep
        prob_param = self.prob_param
        job_func = self.job_func
        # sample_size is a global variable
        data = data_source.sample(sample_size, seed=r)
        with util.ContextTimer() as t:
            tr, te = data.split_tr_te(tr_proportion=tr_proportion, seed=r+21 )
            prob_label = self.prob_label
            logger.info("computing. %s. prob=%s, r=%d,\
                    param=%.3g"%(job_func.__name__, prob_label, r, prob_param))


            job_result = job_func(p, data_source, tr, te, r)

            # create ScalarResult instance
            result = SingleResult(job_result)
            # submit the result to my own aggregator
            self.aggregator.submit_result(result)
            func_name = job_func.__name__
        logger.info("done. ex2: %s, prob=%s, r=%d, param=%.3g. Took: %.3g s "%(func_name,
            prob_label, r, prob_param, t.secs))

        # save result
        fname = '%s-%s-n%d_r%d_p%g_a%.3f_trp%.2f.p' \
                %(prob_label, func_name, sample_size, r, prob_param, alpha,
                        tr_proportion)
        glo.ex_save_result(ex, job_result, prob_label, fname)


# This import is needed so that pickle knows about the class Ex2Job.
# pickle is used when collecting the results from the submitted jobs.
from kgof.ex.ex2_prob_params import Ex2Job
from kgof.ex.ex2_prob_params import job_fssdJ1q_med
from kgof.ex.ex2_prob_params import job_fssdJ5q_med
from kgof.ex.ex2_prob_params import job_fssdJ1q_opt
from kgof.ex.ex2_prob_params import job_fssdJ5q_opt
from kgof.ex.ex2_prob_params import job_fssdJ10q_opt
from kgof.ex.ex2_prob_params import job_fssdJ5p_opt
from kgof.ex.ex2_prob_params import job_fssdJ10p_opt
from kgof.ex.ex2_prob_params import job_fssdJ1q_imq_optv
from kgof.ex.ex2_prob_params import job_fssdJ5q_imq_optv
from kgof.ex.ex2_prob_params import job_fssdJ5q_imqb1_optv
from kgof.ex.ex2_prob_params import job_fssdJ5q_imqb2_optv
from kgof.ex.ex2_prob_params import job_fssdJ5q_imqb3_optv
from kgof.ex.ex2_prob_params import job_fssdJ1q_imq_opt
from kgof.ex.ex2_prob_params import job_fssdJ5q_imq_opt
from kgof.ex.ex2_prob_params import job_fssdJ1q_imq_optbv
from kgof.ex.ex2_prob_params import job_fssdJ5q_imq_optbv
from kgof.ex.ex2_prob_params import job_me_opt
from kgof.ex.ex2_prob_params import job_kstein_med
from kgof.ex.ex2_prob_params import job_kstein_imq
from kgof.ex.ex2_prob_params import job_lin_kstein_med
from kgof.ex.ex2_prob_params import job_mmd_med
from kgof.ex.ex2_prob_params import job_mmd_opt

#--- experimental setting -----
ex = 2

# sample size = n (the training and test sizes are n/2)
sample_size = 1000

# number of test locations / test frequencies J
alpha = 0.05

# training proportion of the FSSD test, MMD-opt test
tr_proportion = 0.2
# repetitions for each parameter setting
reps = 200

method_job_funcs = [ 
        job_fssdJ5q_opt,
        job_fssdJ5q_med, 
        job_kstein_med, 
        job_lin_kstein_med,
        job_mmd_opt,
        job_me_opt,

        #job_fssdJ5q_imq_opt,
        #job_fssdJ5q_imq_optv,
        #job_fssdJ5q_imq_optbv,
        #job_fssdJ5q_imqb1_optv,
        #job_fssdJ5q_imqb2_optv,
        #job_fssdJ5q_imqb3_optv,
        #job_fssdJ10q_opt,
        #job_fssdJ5p_opt,
        #job_fssdJ10p_opt,
        #job_kstein_imq,
        #job_mmd_med,
       ]

# If is_rerun==False, do not rerun the experiment if a result file for the current
# setting of (pi, r) already exists.
is_rerun = False
#---------------------------

def gaussbern_rbm_probs(stds_perturb_B, dx=50, dh=10, n=sample_size):
    """
    Get a sequence of Gaussian-Bernoulli RBM problems.
    We follow the parameter settings as described in section 6 of Liu et al.,
    2016.

    - stds_perturb_B: a list of Gaussian noise standard deviations for perturbing B.
    - dx: observed dimension
    - dh: latent dimension
    """
    probs = []
    for i, std in enumerate(stds_perturb_B):
        with util.NumpySeedContext(seed=i+1000):
            B = np.random.randint(0, 2, (dx, dh))*2 - 1.0
            b = np.random.randn(dx)
            c = np.random.randn(dh)
            p = density.GaussBernRBM(B, b, c)

            if std <= 1e-8:
                B_perturb = B
            else:
                B_perturb = B + np.random.randn(dx, dh)*std
            gb_rbm = data.DSGaussBernRBM(B_perturb, b, c, burnin=2000)

            probs.append((std, p, gb_rbm))
    return probs

def get_pqsource_list(prob_label):
    """
    Return [(prob_param, p, ds) for ... ], a list of tuples
    where 
    - prob_param: a problem parameters. Each parameter has to be a
      scalar (so that we can plot them later). Parameters are preferably
      positive integers.
    - p: a Density representing the distribution p
    - ds: a DataSource, each corresponding to one parameter setting.
        The DataSource generates sample from q.
    """
    sg_ds = [1, 5, 10, 15]
    gmd_ds = [5, 20, 40, 60]
    # vary the mean
    gmd_d10_ms = [0, 0.02, 0.04, 0.06]
    gvinc_d1_vs = [1, 1.5, 2, 2.5] 
    gvinc_d5_vs = [1, 1.5, 2, 2.5]
    gvsub1_d1_vs = [0.1, 0.3, 0.5, 0.7]
    gvd_ds = [1, 5, 10, 15]

    #gb_rbm_dx50_dh10_stds = [0, 0.01, 0.02, 0.03]
    gb_rbm_dx50_dh10_stds = [0, 0.02, 0.04, 0.06]
    #gb_rbm_dx50_dh10_stds = [0]
    gb_rbm_dx50_dh40_stds = [0, 0.01, 0.02, 0.04, 0.06]
    glaplace_ds = [1, 5, 10, 15]
    prob2tuples = { 
            # H0 is true. vary d. P = Q = N(0, I)
            'sg': [(d, density.IsotropicNormal(np.zeros(d), 1),
                data.DSIsotropicNormal(np.zeros(d), 1) ) for d in sg_ds],

            # vary d. P = N(0, I), Q = N( (c,..0), I)
            'gmd': [(d, density.IsotropicNormal(np.zeros(d), 1),
                data.DSIsotropicNormal(np.hstack((1, np.zeros(d-1))), 1) ) 
                for d in gmd_ds
                ],
            # P = N(0, I), Q = N( (m, ..0), I). Vary m
            'gmd_d10_ms': [(m, density.IsotropicNormal(np.zeros(10), 1),
                data.DSIsotropicNormal(np.hstack((m, np.zeros(9))), 1) )
                for m in gmd_d10_ms
                ],
            # d=1. Increase the variance. P = N(0, I). Q = N(0, v*I)
            'gvinc_d1': [(var, density.IsotropicNormal(np.zeros(1), 1),
                data.DSIsotropicNormal(np.zeros(1), var) ) 
                for var in gvinc_d1_vs
                ],
            # d=5. Increase the variance. P = N(0, I). Q = N(0, v*I)
            'gvinc_d5': [(var, density.IsotropicNormal(np.zeros(5), 1),
                data.DSIsotropicNormal(np.zeros(5), var) ) 
                for var in gvinc_d5_vs
                ],
            # d=1. P=N(0,1), Q(0,v). Consider the variance below 1.
            'gvsub1_d1': [(var, density.IsotropicNormal(np.zeros(1), 1),
                data.DSIsotropicNormal(np.zeros(1), var) ) 
                for var in gvsub1_d1_vs
                ],
            # Gaussian variance difference problem. Only the variance 
            # of the first dimenion differs. d varies.
            'gvd': [(d, density.Normal(np.zeros(d), np.eye(d) ), 
                data.DSNormal(np.zeros(d), np.diag(np.hstack((2, np.ones(d-1)))) ))
                for d in gvd_ds],

            # Gaussian Bernoulli RBM. dx=50, dh=10 
            'gbrbm_dx50_dh10': gaussbern_rbm_probs(gb_rbm_dx50_dh10_stds,
                dx=50, dh=10, n=sample_size),

            # Gaussian Bernoulli RBM. dx=50, dh=40
            'gbrbm_dx50_dh40': gaussbern_rbm_probs(gb_rbm_dx50_dh40_stds,
                dx=50, dh=40, n=sample_size),
            
            # p: N(0, I), q: standard Laplace. Vary d
            'glaplace': [(d, density.IsotropicNormal(np.zeros(d), 1), 
                # Scaling of 1/sqrt(2) will make the variance 1.
                data.DSLaplace(d=d, loc=0, scale=1.0/np.sqrt(2))) for d in glaplace_ds],
            }
    if prob_label not in prob2tuples:
        raise ValueError('Unknown problem label. Need to be one of %s'%str(prob2tuples.keys()) )
    return prob2tuples[prob_label]


def run_problem(prob_label):
    """Run the experiment"""
    L = get_pqsource_list(prob_label)
    prob_params, ps, data_sources = zip(*L)
    # make them lists 
    prob_params = list(prob_params)
    ps = list(ps)
    data_sources = list(data_sources)

    # ///////  submit jobs //////////
    # create folder name string
    #result_folder = glo.result_folder()
    from kgof.config import expr_configs
    tmp_dir = expr_configs['scratch_path']
    foldername = os.path.join(tmp_dir, 'kgof_slurm', 'e%d'%ex)
    logger.info("Setting engine folder to %s" % foldername)

    # create parameter instance that is needed for any batch computation engine
    logger.info("Creating batch parameter instance")
    batch_parameters = BatchClusterParameters(
        foldername=foldername, job_name_base="e%d_"%ex, parameter_prefix="")

    # Use the following line if Slurm queue is not used.
    #engine = SerialComputationEngine()
    engine = SlurmComputationEngine(batch_parameters)
    #engine = SlurmComputationEngine(batch_parameters, partition='wrkstn,compute')
    n_methods = len(method_job_funcs)
    # repetitions x len(prob_params) x #methods
    aggregators = np.empty((reps, len(prob_params), n_methods ), dtype=object)
    for r in range(reps):
        for pi, param in enumerate(prob_params):
            for mi, f in enumerate(method_job_funcs):
                # name used to save the result
                func_name = f.__name__
                fname = '%s-%s-n%d_r%d_p%g_a%.3f_trp%.2f.p' \
                    %(prob_label, func_name, sample_size, r, param, alpha,
                            tr_proportion)
                if not is_rerun and glo.ex_file_exists(ex, prob_label, fname):
                    logger.info('%s exists. Load and return.'%fname)
                    job_result = glo.ex_load_result(ex, prob_label, fname)

                    sra = SingleResultAggregator()
                    sra.submit_result(SingleResult(job_result))
                    aggregators[r, pi, mi] = sra
                else:
                    # result not exists or rerun

                    # p: an UnnormalizedDensity object
                    p = ps[pi]
                    job = Ex2Job(SingleResultAggregator(), p, data_sources[pi],
                            prob_label, r, f, param)
                    agg = engine.submit_job(job)
                    aggregators[r, pi, mi] = agg

    # let the engine finish its business
    logger.info("Wait for all call in engine")
    engine.wait_for_all()

    # ////// collect the results ///////////
    logger.info("Collecting results")
    job_results = np.empty((reps, len(prob_params), n_methods), dtype=object)
    for r in range(reps):
        for pi, param in enumerate(prob_params):
            for mi, f in enumerate(method_job_funcs):
                logger.info("Collecting result (%s, r=%d, param=%.3g)" %
                        (f.__name__, r, param))
                # let the aggregator finalize things
                aggregators[r, pi, mi].finalize()

                # aggregators[i].get_final_result() returns a SingleResult instance,
                # which we need to extract the actual result
                job_result = aggregators[r, pi, mi].get_final_result().result
                job_results[r, pi, mi] = job_result

    #func_names = [f.__name__ for f in method_job_funcs]
    #func2labels = exglobal.get_func2label_map()
    #method_labels = [func2labels[f] for f in func_names if f in func2labels]

    # save results 
    results = {'job_results': job_results, 'prob_params': prob_params, 
            'alpha': alpha, 'repeats': reps, 
            'ps': ps,
            'list_data_source': data_sources, 
            'tr_proportion': tr_proportion,
            'method_job_funcs': method_job_funcs, 'prob_label': prob_label,
            'sample_size': sample_size, 
            }
    
    # class name 
    fname = 'ex%d-%s-me%d_n%d_rs%d_pmi%g_pma%g_a%.3f_trp%.2f.p' \
        %(ex, prob_label, n_methods, sample_size, reps, min(prob_params),
                max(prob_params), alpha, tr_proportion)

    glo.ex_save_result(ex, results, fname)
    logger.info('Saved aggregated results to %s'%fname)


def main():
    if len(sys.argv) != 2:
        print('Usage: %s problem_label'%sys.argv[0])
        sys.exit(1)
    prob_label = sys.argv[1]

    run_problem(prob_label)

if __name__ == '__main__':
    main()




import time
import warnings
from typing import Any, Dict, Optional

import torch
from pyro.infer.mcmc import HMC, NUTS
from sbi.mcmc.mcmc import MCMC
from sbi.mcmc.slice import Slice

import sbibm
from sbibm.algorithms.pyro.utils.tensorboard import (
    tb_acf,
    tb_ess,
    tb_make_hook_fn,
    tb_marginals,
    tb_posteriors,
    tb_r_hat,
)
from sbibm.tasks.task import Task
from sbibm.utils.tensorboard import tb_make_writer, tb_plot_posterior


def run(
    task: Task,
    num_samples: int,
    num_observation: Optional[int] = None,
    observation: Optional[torch.Tensor] = None,
    num_chains: int = 10,
    num_warmup: int = 10000,
    kernel: str = "slice",
    kernel_parameters: Optional[Dict[str, Any]] = None,
    thinning: int = 1,
    diagnostics: bool = True,
    available_cpu: int = 1,
    mp_context: str = "fork",
    jit_compile: bool = False,
    automatic_transforms_enabled: bool = True,
    initial_params: Optional[torch.Tensor] = None,
    **kwargs: Any,
) -> torch.Tensor:
    """Runs MCMC using Pyro on potential function

    Produces `num_samples` while accounting for warmup (burn-in) and thinning.

    Note that the actual number of simulations is not controlled for with MCMC since
    algorithms are only used as a reference method in the benchmark. 

    MCMC is run on the potential function, which returns the unnormalized
    negative log posterior probability. Note that this requires a tractable likelihood.
    Pyro is used to automatically construct the potential function.

    Args:
        task: Task instance
        num_samples: Number of samples to generate from posterior
        num_observation: Observation number to load, alternative to `observation`
        observation: Observation, alternative to `num_observation`
        num_chains: Number of chains
        num_warmup: Warmup steps, during which parameters of the sampler are adapted.
            Warmup samples are not returned by the algorithm.
        kernel: HMC, NUTS, or Slice
        kernel_parameters: Parameters passed to kernel
        thinning: Amount of thinning to apply, in order to avoid drawing
            correlated samples from the chain
        diagnostics: Flag for diagnostics
        available_cpu: Number of CPUs used to parallelize chains
        mp_context: multiprocessing context, only fork might work
        jit_compile: Just-in-time (JIT) compilation, can yield significant speed ups
        automatic_transforms_enabled: Whether or not to use automatic transforms
        initial_params: Parameters to initialize at

    Returns:
        Samples from posterior
    """
    assert not (num_observation is None and observation is None)
    assert not (num_observation is not None and observation is not None)

    tic = time.time()
    log = sbibm.get_logger(__name__)

    hook_fn = None
    if diagnostics:
        log.info(f"MCMC sampling for observation {num_observation}")
        tb_writer, tb_close = tb_make_writer(
            logger=log,
            basepath=f"tensorboard/pyro_{kernel.lower()}/observation_{num_observation}",
        )
        hook_fn = tb_make_hook_fn(tb_writer)

    if "num_simulations" in kwargs:
        warnings.warn(
            "`num_simulations` was passed as a keyword but will be ignored, see docstring for more info."
        )

    # Prepare model and transforms
    conditioned_model = task._get_pyro_model(
        num_observation=num_observation, observation=observation
    )
    transforms = task._get_transforms(
        num_observation=num_observation,
        observation=observation,
        automatic_transforms_enabled=automatic_transforms_enabled,
    )

    kernel_parameters = kernel_parameters if kernel_parameters is not None else {}
    kernel_parameters["jit_compile"] = jit_compile
    kernel_parameters["transforms"] = transforms
    log.info(
        "Using kernel: {name}({parameters})".format(
            name=kernel,
            parameters=",".join([f"{k}={v}" for k, v in kernel_parameters.items()]),
        )
    )
    if kernel.lower() == "nuts":
        mcmc_kernel = NUTS(model=conditioned_model, **kernel_parameters)

    elif kernel.lower() == "hmc":
        mcmc_kernel = HMC(model=conditioned_model, **kernel_parameters)

    elif kernel.lower() == "slice":
        mcmc_kernel = Slice(model=conditioned_model, **kernel_parameters)

    else:
        raise NotImplementedError

    if initial_params is not None:
        site_name = "parameters"
        initial_params = {site_name: transforms[site_name](initial_params)}
    else:
        initial_params = None

    mcmc_parameters = {
        "num_chains": num_chains,
        "num_samples": thinning * num_samples,
        "warmup_steps": num_warmup,
        "available_cpu": available_cpu,
        "initial_params": initial_params,
    }
    log.info(
        "Calling MCMC with: MCMC({name}_kernel, {parameters})".format(
            name=kernel,
            parameters=",".join([f"{k}={v}" for k, v in mcmc_parameters.items()]),
        )
    )

    mcmc = MCMC(mcmc_kernel, hook_fn=hook_fn, **mcmc_parameters)
    mcmc.run()

    toc = time.time()
    log.info(f"Finished MCMC after {toc-tic:.3f} seconds")
    log.info(f"Automatic transforms {mcmc.transforms}")

    log.info(f"Apply thinning of {thinning}")
    mcmc._samples = {"parameters": mcmc._samples["parameters"][:, ::thinning, :]}

    num_samples_available = (
        mcmc._samples["parameters"].shape[0] * mcmc._samples["parameters"].shape[1]
    )
    if num_samples_available < num_samples:
        warnings.warn("Some samples will be included multiple times")
        samples = mcmc.get_samples(num_samples=num_samples, group_by_chain=False)[
            "parameters"
        ].squeeze()
    else:
        samples = mcmc.get_samples(group_by_chain=False)["parameters"].squeeze()
        idx = torch.randperm(samples.shape[0])[:num_samples]
        samples = samples[idx, :]

    assert samples.shape[0] == num_samples

    if diagnostics:
        mcmc.summary()
        tb_ess(tb_writer, mcmc)
        tb_r_hat(tb_writer, mcmc)
        tb_marginals(tb_writer, mcmc)
        tb_acf(tb_writer, mcmc)
        tb_posteriors(tb_writer, mcmc)
        tb_plot_posterior(tb_writer, samples, tag="posterior/final")
        tb_close()

    return samples



import logging
import math
from typing import Any, Dict, Optional

import torch
import torch.nn as nn
from sbi import inference as inference
from sbi.inference.posteriors.likelihood_based_posterior import LikelihoodBasedPosterior
from torch import Tensor

from sbibm.algorithms.sbi.utils import (
    wrap_posterior,
    wrap_prior_dist,
    wrap_simulator_fn,
)
from sbibm.tasks.task import Task


class SynthLikNet(nn.Module):
    def __init__(self, simulator, num_simulations_per_step=100, diag_eps=0.0):
        self.simulator = simulator
        self.num_simulations_per_step = num_simulations_per_step
        self.diag_eps = diag_eps

    def forward(self, *args, **kwargs):
        raise NotImplementedError

    def log_prob(self, inputs: Tensor, context=Optional[Tensor]) -> Tensor:
        thetas = context
        observation = inputs

        log_probs = []
        for i in range(thetas.shape[0]):
            xs = self.simulator(
                thetas[i, :].reshape(1, -1).repeat(self.num_simulations_per_step, 1)
            )

            # Estimate mean and covariance of MVN
            m = torch.mean(xs, dim=0)
            xm = xs - m
            S = torch.matmul(xm.T, xm) / xs.shape[0]
            S = S + self.diag_eps * torch.eye(xs.shape[1])

            # Score
            dist = torch.distributions.MultivariateNormal(loc=m, covariance_matrix=S)
            log_probs.append(dist.log_prob(observation[i, :].reshape(1, -1)))

        return torch.cat(log_probs)

    def sample(self, *args, **kwargs):
        raise NotImplementedError

    def train(self, *args, **kwargs):
        pass

    def eval(self, *args, **kwargs):
        pass


def run(
    task: Task,
    num_samples: int,
    num_simulations: int,
    num_simulations_per_step: int = 100,
    num_observation: Optional[int] = None,
    observation: Optional[torch.Tensor] = None,
    automatic_transforms_enabled: bool = False,
    mcmc_method: str = "slice_np",
    mcmc_parameters: Dict[str, Any] = {},
    diag_eps: float = 0.0,
) -> (torch.Tensor, int, Optional[torch.Tensor]):
    """Runs (S)NLE from `sbi`

    Args:
        task: Task instance
        num_observation: Observation number to load, alternative to `observation`
        observation: Observation, alternative to `num_observation`
        num_samples: Number of samples to generate from posterior
        num_simulations: Simulation budget
        num_simulations_per_step: Number of simulations per MCMC step
        automatic_transforms_enabled: Whether to enable automatic transforms
        mcmc_method: MCMC method
        mcmc_parameters: MCMC parameters
        diag_eps: Epsilon applied to diagonal

    Returns:
        Samples from posterior, number of simulator calls, log probability of true params if computable
    """
    assert not (num_observation is None and observation is None)
    assert not (num_observation is not None and observation is not None)

    log = logging.getLogger(__name__)

    log.info(f"Running SL")

    prior = task.get_prior_dist()
    if observation is None:
        observation = task.get_observation(num_observation)

    simulator = task.get_simulator()

    transforms = task._get_transforms(automatic_transforms_enabled)["parameters"]
    prior = wrap_prior_dist(prior, transforms)
    simulator = wrap_simulator_fn(simulator, transforms)

    likelihood_estimator = SynthLikNet(
        simulator=simulator,
        num_simulations_per_step=num_simulations_per_step,
        diag_eps=diag_eps,
    )

    posterior = LikelihoodBasedPosterior(
        method_family="snle",
        neural_net=likelihood_estimator,
        prior=prior,
        x_shape=observation.shape,
        mcmc_parameters=mcmc_parameters,
    )

    posterior.set_default_x(observation)

    posterior = wrap_posterior(posterior, transforms)

    # assert simulator.num_simulations == num_simulations

    samples = posterior.sample((num_samples,)).detach()

    return samples, simulator.num_simulations, None
